import pandas as pd
import numpy as np
from scipy.interpolate import RectBivariateSpline
from pathlib import Path


class GVLookup:
    def __init__(self, path: str = None):
        """Initialize with CSV file containing GV lookup table."""

        if path is None:
            this_file_dir = Path(__file__).parent
            csv_path = this_file_dir / "assets" / "RC_2020.csv"
        else:
            csv_path = Path(path)
        self.df = pd.read_csv(csv_path, index_col=0)
        self.lats = self.df.index.astype(float).values
        self.lons = self.df.columns.astype(float).values
        self.values = self.df.values.astype(float)

        # Note: RectBivariateSpline requires strictly increasing coordinates
        self.interpolator = RectBivariateSpline(
            self.lats[::-1],  # Reverse for ascending order
            self.lons,
            self.values[::-1],  # Reverse to match lats
            s=0,
        )

    def get_gv(self, lat, lon):
        """
        Get GV value for given lat/lon coordinates.

        Args:
            lat: Latitude (-90 to 90)
            lon: Longitude (-180 to 180)

        Returns:
            Interpolated GV value
        """
        lat = np.clip(lat, -90, 90)
        lon = np.clip(lon, -180, 180)

        return round(float(self.interpolator(lat, lon)[0, 0]), 2)



import pandas as pd
import numpy as np
from neptoon.logging import get_logger
from neptoon.columns import ColumnInfo
from .correction_classes import (
    Correction,
    CorrectionType,
    CorrectionTheory,
    IncomingIntensityCorrectionZreda2012,
    IncomingIntensityCorrectionHawdon2014,
    IncomingIntensityCorrectionMcJannetDesilets2023,
    HumidityCorrectionRosolem2013,
    PressureCorrectionDesiletsZreda2003,
    PressureCorrectionTiradoBueno2021,
    PressureCorrectionDesilets2021,
    AboveGroundBiomassCorrectionBaatz2015,
    AboveGroundBiomassCorrectionMorris2024,
)

core_logger = get_logger()


def calculate_poisson_uncertainty():
    pass


class CorrectionBuilder:
    """
    Staging place for the corrections as they are built. First a user
    adds a check using the add_check method.

    Parameters
    ----------

    corrections : dict
        dictionary which contains the corrections. The key is the
        correction_type assigned in each correction, the value is the
        correction itself.

    """

    def __init__(self):
        self.corrections = {}

    def add_correction(self, correction: Correction):
        """
        Adds a correction to the CorrectionBuilder

        Parameters
        ----------
        correction : Correction
            A Correction object
        """

        if isinstance(correction, Correction):
            correction_type = correction.correction_type
            self.corrections[correction_type] = correction
        else:
            message = f"{correction} is not a correction"
            core_logger.error(message)

    def remove_correction_by_type(self, correction_type: str):
        """
        Removes a correction from the CorrectionBuilder based on its
        type

        Parameters
        ----------
        correction_type : str
            The type of correction to be removed
        """
        if correction_type in self.corrections:
            self.corrections.pop(correction_type)
        else:
            raise ValueError(
                f"Correction type '{correction_type}' not found in the builder."
            )

    def get_corrections(self):
        """
        Returns the corrections stored in the builder
        """
        return self.corrections.values()


class CorrectNeutrons:
    """
    CorrectNeutrons class handles correcting neutrons for additional
    influences beyond soil moisture. It takes in a crns_data_frame which
    is a pd.DataFrame which stores the required data.

    Methods are available for staging a series of corrections which are
    applied to remove additional influences on the neutron signal. A
    user can add corrections individually, or create a CorrectionBuilder
    class seperately that has been pre-compiled and inject that into the
    CorrectNeutrons instance.

    Once the CorrectionBuilder has been appropriately staged with
    desired corrections, the correct_neutrons method will apply each
    correction, record the correction factor and create a corrected
    epithermal neutron count column with the correction factors applied.
    """

    def __init__(
        self,
        crns_data_frame: pd.DataFrame,
        correction_builder: CorrectionBuilder,
    ):
        """
        Attributes for using the CorrectNeutrons class

        Parameters
        ----------
        crns_data_frame : pd.DataFrame
            A DataFrame which contains the appropriate information to
            apply corrections.
        correction_builder : CorrectionBuilder
            Staging area for corrections. Can be built or supplied
            completed.
        """
        self._crns_data_frame = crns_data_frame
        self._correction_builder = correction_builder
        self._correction_columns = []

    @property
    def crns_data_frame(self):
        return self._crns_data_frame

    @crns_data_frame.setter
    def crns_data_frame(self, df: pd.DataFrame):
        # TODO add checks that it is df
        self._crns_data_frame = df

    @property
    def correction_builder(self):
        return self._correction_builder

    @correction_builder.setter
    def correction_builder(self, builder: CorrectionBuilder):
        if isinstance(builder, CorrectionBuilder):
            self._correction_builder = builder
        else:
            message = f"It appears {builder} is not a CorrectionBuilder object"
            core_logger.error(message)
            raise AttributeError

    @property
    def correction_columns(self):
        return self._correction_columns

    def add_correction(self, new_correction: Correction):
        """
        Add an invidual correction to the CorrectionBuilder

        Parameters
        ----------
        new_correction : Correction
            The new correction to apply
        """
        self.correction_builder.add_correction(new_correction)

    def add_correction_builder(
        self, new_correction_builder: CorrectionBuilder
    ):
        """
        Add a whole correction builder. Useful if a correction builder
        has been built somewhere in the code and you want to read it in.

        Parameters
        ----------
        new_correction_builder : CorrectionBuilder
            A pre-compiled correction builder.
        """
        self.correction_builder = new_correction_builder

    def create_correction_factors(self, df: pd.DataFrame):
        """
        Cycles through all the corrections in the CorrectionBuilder and
        applies them to the DataFrame. Returns the DataFrame with additional
        columns.

        Parameters
        ----------
        df : pd.DataFrame
            DataFrame which is prepared for correction.

        Returns
        -------
        pd.DataFrame
            DataFrame with additional columns applied during correction.
        """
        for correction in self.correction_builder.get_corrections():
            df = correction.apply(df)
            correction_column_name = (
                correction.get_correction_factor_column_name()
            )
            self.correction_columns.append(correction_column_name)

        return df

    def create_corrected_neutron_column(self, df):
        """
        Calculates the corrected neutron count rate after applying all
        the corrections.

        NOTE:

        Parameters
        ----------
        df : pd.DataFrame
            DataFrame with the corrections applied and recorded in the
            columns

        Returns
        -------
        pd.DataFrame
            DataFrame with the corrected epithermal neutron count
            recorded in a column
        """
        df[str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT)] = df[
            str(ColumnInfo.Name.EPI_NEUTRON_COUNT_FINAL)
        ]
        for column_name in self.correction_columns:
            df[str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT)] = (
                df[str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT)]
                * df[column_name]
            )
        return df

    def create_corrected_neutron_uncertainty_column(self, df: pd.DataFrame):
        """
        Creates corrected epithermal neutron uncertainty data

        Parameters
        ----------
        df : pd.DataFrame
            DataFrame
        """
        df[str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_UNCERTAINTY)] = df[
            str(ColumnInfo.Name.RAW_EPI_NEUTRON_COUNT_UNCERTAINTY)
        ]
        for column_name in self.correction_columns:
            df[
                str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_UNCERTAINTY)
            ] = (
                df[
                    str(
                        ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_UNCERTAINTY
                    )
                ]
                * df[column_name]
            )
        return df

    def correct_neutrons(self):
        """
        Corrects neutrons using the CorrectionBuilder. Returns the
        DataFrame.

        Returns
        -------
        df: pd.DataFrame
            DataFrame returned with additional columns.
        """
        df = self.create_correction_factors(self.crns_data_frame)
        df = self.create_corrected_neutron_column(df)
        df = self.create_corrected_neutron_uncertainty_column(df)

        return df


class CorrectionFactory:
    """
    Used inside the CRNSDataHub when selecting which corrections to
    apply. Creating a correction involves providing a CorrectionType and
    (optionally) a CorrectionTheory. The factory will then use
    information from the SiteInformation object to create a Correction
    with the appropriate information provided. Additionally a user can
    register a custom correction for use in processing.
    """

    def __init__(self):
        self.custom_corrections = {}

    def create_correction(
        self,
        correction_type: CorrectionType,
        correction_theory: CorrectionTheory = None,
    ):
        """
        Creates a particular Correction. CorrectionType and
        CorrectionTheory enums are used for selection. If
        CorrectionTheory is left empty the default correction is
        selected.

        Parameters
        ----------
        correction_type : CorrectionType
            The correction type to stage can be:
                - CorrectionType.INCOMING_INTENSITY
                - CorrectionType.ABOVE_GROUND_BIOMASS
                - CorrectionType.PRESSURE
                - CorrectionType.HUMIDITY
                - CorrectionType.CUSTOM
        correction_theory : CorrectionTheory, optional
            The theory to apply leave blank to use the default format,
            otherwise see CorrectionTheory for options, by default None

        Returns
        -------
        Correction
            Returns a correction object with the site specific values
            read in from the SiteInfo class
        """
        if (correction_type, correction_theory) in self.custom_corrections:
            return self.custom_corrections[
                (correction_type, correction_theory)
            ](
                correction_type=correction_type,
            )
        if correction_type == CorrectionType.ABOVE_GROUND_BIOMASS:
            return self.create_biomass_correction(
                correction_theory=correction_theory
            )
        if correction_type == CorrectionType.INCOMING_INTENSITY:
            return self.create_intensity_correction(
                correction_theory=correction_theory
            )
        if correction_type == CorrectionType.PRESSURE:
            return self.create_pressure_correction(
                correction_theory=correction_theory
            )
        if correction_type == CorrectionType.HUMIDITY:
            return self.create_humidity_correction(
                correction_theory=correction_theory
            )

    def create_intensity_correction(self, correction_theory: CorrectionTheory):
        """
        Internal method for selecting the incoming neutron intensity
        correction to use. If no CorrectionTheory supplied it will use
        the default.

        Parameters
        ----------
        correction_theory : CorrectionTheory
            The CorrectionTheory to use

        Returns
        -------
        Correction
            Intensity correction with values filled in.
        """
        if correction_theory is None:
            return IncomingIntensityCorrectionHawdon2014()
        elif correction_theory == CorrectionTheory.ZREDA_2012:
            return IncomingIntensityCorrectionZreda2012()
        elif correction_theory == CorrectionTheory.HAWDON_2014:
            return IncomingIntensityCorrectionHawdon2014()
        elif correction_theory == CorrectionTheory.MCJANNET_DESILETS_2023:
            return IncomingIntensityCorrectionMcJannetDesilets2023()

    def create_biomass_correction(self, correction_theory: CorrectionTheory):
        """
        Internal method for selecting the above ground biomass
        correction to use. If no CorrectionTheory supplied it will use
        the default.

        NOTE:

        Parameters
        ----------
        correction_theory : CorrectionTheory
            The CorrectionTheory to use

        Returns
        -------
        Correction
            Above Ground Biomass correction with values filled in.
        """
        if correction_theory is None:
            return AboveGroundBiomassCorrectionBaatz2015()
        elif correction_theory == CorrectionTheory.BAATZ_2015:
            return AboveGroundBiomassCorrectionBaatz2015()
        elif correction_theory == CorrectionTheory.MORRIS_2024:
            return AboveGroundBiomassCorrectionMorris2024()

    def create_pressure_correction(self, correction_theory: CorrectionTheory):
        """
        Internal method for selecting the correct pressure correction to
        use.

        Parameters
        ----------
        correction_theory : CorrectionTheory
            The CorrectionTheory to apply

        Returns
        -------
        Correction
            Returns a pressure Correction with values inserted.
        """
        if correction_theory is None:
            return PressureCorrectionTiradoBueno2021()
        elif correction_theory is CorrectionTheory.TIRADO_BUENO_2021:
            return PressureCorrectionTiradoBueno2021()
        elif correction_theory is CorrectionTheory.DESILETS_2021:
            return PressureCorrectionDesilets2021()
        elif correction_theory is CorrectionTheory.DESILETS_ZREDA_2003:
            return PressureCorrectionDesiletsZreda2003()

    def create_humidity_correction(
        self, correction_theory: CorrectionTheory = None
    ):
        """
        Internal method for selecting the correct humidity correction to
        use

        Parameters
        ----------
        correction_theory : CorrectionTheory
            The CorrectionTheory to apply

        Returns
        -------
        Correction
            Returns the Correction
        """
        if correction_theory is None:
            return HumidityCorrectionRosolem2013()
        elif correction_theory == CorrectionTheory.ROSOLEM_2013:
            return HumidityCorrectionRosolem2013()

    def register_custom_correction(
        self, correction_type: CorrectionType, theory: str, correction_class
    ):
        """
        Used to register a custom correction theory. Without
        registration the correction cannot be used.

        The correction_class must be an object which inherits the
        Correction class and follows the same principles as other
        Correction type objects. This means it requires an apply method,
        which takes in a dataframe and returns a dataframe along with
        updates calulcated along the way.

        The key difference of a custom Correction class is that it must
        take a SiteInformation object. The logic under apply() will then
        use this SiteInformation object to collect relevant information
        for processing.

        Parameters
        ----------
        correction_type : CorrectionType
            The type of correction being registered
        theory : str
            A name for the custom correction
        correction_class : Correction
            A custom Correction class. See documentation on how to make
            this.

        Raises
        ------
        ValueError
            If the correction class is not a Correction type object it
            throws an error
        """
        if not issubclass(correction_class, Correction):
            message = "Custom correction must inherit from Correction class"
            core_logger.error(message)
            raise ValueError(message)
        self.custom_corrections[(correction_type, theory)] = correction_class



import pytest
from unittest.mock import patch
from neptoon.columns import ColumnInfo


@pytest.fixture(autouse=True)
def mock_logging(request):
    """
    Turn of logging for tests.
    """
    if "test_logging" in request.keywords:
        yield
    else:
        with patch("logging.Logger") as MockLogger:
            yield MockLogger


@pytest.fixture()
def reset_column_info(autouse=True):
    """
    Automatically resets ColumnInfo labels before runnning test.
    Important for tests related to ColumnInfo renaming.
    """
    ColumnInfo.reset_labels()
    yield
    ColumnInfo.reset_labels()


# Marker for tests that change ColumnInfo
pytest.mark.reset_columns = pytest.mark.usefixtures("reset_column_info")



from .general_utils import (
    validate_and_convert_file_path,
    parse_resolution_to_timedelta,
    find_temporal_resolution_seconds,
    validate_df,
    validate_timestamp_index,
    timedelta_to_freq_str,
    is_resolution_greater_than,
    recalculate_neutron_uncertainty,
)



from neptoon.data_prep.cutoff_rigidity_lookup import GVLookup


def test_gv_get_interpolate():
    lat = 10
    lon = 10
    lookup = GVLookup()
    gv = lookup.get_gv(lat=lat, lon=lon)
    assert gv == 15.06


def test_gv_get_exact():
    lat = 45
    lon = 0
    lookup = GVLookup()
    gv = lookup.get_gv(lat=lat, lon=lon)
    assert gv == 4.97



import pandas as pd
from dataclasses import dataclass
from neptoon.config.configuration_input import SensorInfo
from neptoon.visulisation.figures import (
    make_nmdb_data_figure,
    soil_moisture_coloured_figure,
    soil_moisture_figure_uncertainty,
    uncorr_and_corr_neutrons_figure,
    atmospheric_conditions_figure,
    correction_factors_figure,
)
from neptoon.columns import ColumnInfo
from typing import List, Optional
from enum import Enum
from magazine import Magazine
from pathlib import Path
import tempfile
import atexit
import shutil
import secrets
import warnings

warnings.filterwarnings(
    "ignore",
    message=".*FigureCanvasAgg is non-interactive.*",
    category=UserWarning,
)


class FigureTopic(Enum):
    NMDB = "nmdb"
    NEUTRONS = "neutrons"
    SOIL_MOISTURE = "soil_moisture"
    ATMOSPHERIC = "atmospheric"


@dataclass
class FigureMetadata:
    """
    Key information on each figure created.

    topic : FigureTopic
        The topic the figure is related to
    description : str
        A brief description about the figure
    required_columns: List[str]
        The columns required for this particular figure.
    method : callable
        The method (found in FigureHandler) which will be called to
        produce this figure.
    """

    topic: FigureTopic
    description: str
    required_columns: List[str]
    method: callable


@dataclass
class TempFigure:
    """
    Tracks a temporary figure file

    name : str
        Name of the figure

    """

    name: str
    path: Path
    topic: str


class TempFigureHandler:
    """
    Handles the creation and storage of figures made with the
    FigureHandler by creating a temp directory, recording paths for each
    figure, and removing temp files upon program exit.
    """

    def __init__(
        self,
        site_id: str,
    ):
        self.site_id = site_id
        self._temp_dir = Path(
            tempfile.mkdtemp(prefix="neptoon_figures_"),
            self.site_id,
        )
        self.figures: List[TempFigure] = []
        atexit.register(self.cleanup)

    def store_figure(self, name: str, topic: FigureTopic):
        """
        Stores metadata of temporary figure storage

        Parameters
        ----------
        name : str
            name of the figure being stored
        topic : FigureTopic
            Topic the figure is part of (e.g., NMDB, or Soil)

        Returns
        -------
        Path
            Path of the temp figure
        """
        temp_path = self._temp_dir / f"{name}.png"
        self.figures.append(
            TempFigure(
                name=name,
                path=temp_path,
                topic=topic,
            )
        )
        return temp_path

    def cleanup(self):
        """Remove temporary directory and files"""
        if self._temp_dir.exists():
            shutil.rmtree(self._temp_dir)

    def get_figures(self, topic: Optional[FigureTopic] = None):
        """
        Collect the figures that are saved

        Parameters
        ----------
        topic : FigureTopic
            Topic the figure is part of (e.g., NMDB, or Soil)

        Returns
        -------
        List[TempFigure]
            A list of TempFigure objects with location of save files.
        """
        if topic is None:
            return self.figures
        return [fig for fig in self.figures if fig.topic == topic]


class FigureHandler:
    """
    Class to manage creating figures when data has been processed using the CRNSDataHub.
    """

    _figure_registry = {
        "nmdb_incoming_radiation": FigureMetadata(
            topic=FigureTopic.NMDB,
            description="NMDB incoming cosmic radiation plot",
            method="_nmdb_incoming_radiation",
            required_columns=[
                str(ColumnInfo.Name.INCOMING_NEUTRON_INTENSITY),
                str(ColumnInfo.Name.REFERENCE_INCOMING_NEUTRON_VALUE),
            ],
        ),
        "correction_factors_figure": FigureMetadata(
            topic=FigureTopic.NEUTRONS,
            description="Correction factors applied to raw neutron counts",
            method="_neutron_correction_factors",
            required_columns=[],  # empty for flexibility
        ),
        "neutron_counts_corr_uncorr": FigureMetadata(
            topic=FigureTopic.NEUTRONS,
            description="Corrected and uncorrected count rates at the sensor",
            method="_uncorr_and_corr_neutrons_figure",
            required_columns=[
                str(ColumnInfo.Name.EPI_NEUTRON_COUNT_FINAL),
                str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_FINAL),
            ],
        ),
        "soil_moisture_uncertainty": FigureMetadata(
            topic=FigureTopic.SOIL_MOISTURE,
            description="Soil moisture time series with uncertainty bounds",
            method="_soil_moisture_uncertainty",
            required_columns=[
                str(ColumnInfo.Name.SOIL_MOISTURE_VOL_FINAL),
                str(ColumnInfo.Name.SOIL_MOISTURE_UNCERTAINTY_VOL_LOWER),
                str(ColumnInfo.Name.SOIL_MOISTURE_UNCERTAINTY_VOL_UPPER),
            ],
        ),
        # "soil_moisture_coloured": FigureMetadata(
        #     topic=FigureTopic.SOIL_MOISTURE,
        #     description="Soil moisture time series with colour filling",
        #     method="_soil_moisture_colour",
        #     required_columns=[str(ColumnInfo.Name.SOIL_MOISTURE_FINAL)],
        # ),
        "atmospheric_variables": FigureMetadata(
            topic=FigureTopic.ATMOSPHERIC,
            description="Atmospheric variables.",
            method="_atmospheric_variables",
            required_columns=[
                str(ColumnInfo.Name.AIR_PRESSURE),
                str(ColumnInfo.Name.AIR_TEMPERATURE),
                str(ColumnInfo.Name.AIR_RELATIVE_HUMIDITY),
                str(ColumnInfo.Name.ABSOLUTE_HUMIDITY),
            ],
        ),
    }

    def __init__(
        self,
        data_frame: pd.DataFrame,
        sensor_info: SensorInfo,
        create_all: bool = False,
        ignore_sections: List = None,
        selected_figures: List[str] = None,
        show_figures: bool = False,
        backend: str = "Agg",
    ):

        self.data_frame = data_frame
        self.sensor_info = sensor_info
        self.site_id = secrets.token_hex(3)
        self.temp_handler = TempFigureHandler(site_id=self.site_id)
        self.create_all = create_all
        self.ignore_sections = (
            ignore_sections if ignore_sections is not None else []
        )
        self.selected_figures = (
            selected_figures if selected_figures is not None else []
        )
        self.show_figures = show_figures
        self.backend = backend

    def _validate_required_columns(self, metadata: FigureMetadata):
        """
        Validates that the required columns are present in the data
        frame

        Parameters
        ----------
        metadata : FigureMetadata
            An instance of FigureMetadata

        Raises
        ------
        ValueError
            When missing columns necessary for processing
        """

        missing_cols = [
            col
            for col in metadata.required_columns
            if col not in self.data_frame.columns
        ]
        if missing_cols:
            raise ValueError(
                f"DataFrame is missing required columns: {missing_cols}"
            )

    def _resolve_method(self, metadata: FigureMetadata):
        """
        Resolves the method string to an actual callable method.

        """
        method_name = metadata.method
        if not hasattr(self, method_name):
            raise AttributeError(
                f"Method {method_name} not found in {self.__class__.__name__}"
            )
        return getattr(self, method_name)

    @Magazine.reporting_figure(topic="NMDB")
    def _nmdb_incoming_radiation(self):
        """
        Implements nmdb figure 1
        """
        temp_path = self.temp_handler.store_figure(
            name="nmdb_incoming_radiation",
            topic=FigureTopic.NMDB,
        )
        reference_value = self.data_frame[
            str(ColumnInfo.Name.REFERENCE_INCOMING_NEUTRON_VALUE)
        ].iloc[0]

        make_nmdb_data_figure(
            data_frame=self.data_frame,
            # nmdb_station_name=self.sensor_info.name,
            reference_value=reference_value,
            incoming_neutron_col_name=str(
                ColumnInfo.Name.INCOMING_NEUTRON_INTENSITY
            ),
            show=self.show_figures,
            backend=self.backend,
            save_location=temp_path,
        )

    @Magazine.reporting_figure(topic="Neutron Correction")
    def _uncorr_and_corr_neutrons_figure(self):
        """
        Implements uncorrected and corrected neutrons figures
        """
        temp_path = self.temp_handler.store_figure(
            name="neutron_counts_corr_uncorr",
            topic=FigureTopic.NEUTRONS,
        )
        uncorr_and_corr_neutrons_figure(
            data_frame=self.data_frame,
            station_name=self.sensor_info.name,
            raw_neutron_col_name=str(ColumnInfo.Name.EPI_NEUTRON_COUNT_FINAL),
            corr_neutron_col_name=str(
                ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_FINAL
            ),
            show=self.show_figures,
            backend=self.backend,
            save_location=temp_path,
        )

    @Magazine.reporting_figure(topic="Soil Moisture")
    def _soil_moisture_uncertainty(self):
        """
        Implements soil moisture with uncertainty figure.
        """
        temp_path = self.temp_handler.store_figure(
            name="soil_moisture_uncertainty",
            topic=FigureTopic.SOIL_MOISTURE,
        )

        sm_max = (
            self.data_frame[str(ColumnInfo.Name.SOIL_MOISTURE_VOL_FINAL)].max()
        ) * 1.1
        sm_min = 0
        sm_range = (sm_min, sm_max)

        soil_moisture_figure_uncertainty(
            data_frame=self.data_frame,
            station_name=self.sensor_info.name,
            soil_moisture_col=str(ColumnInfo.Name.SOIL_MOISTURE_VOL_FINAL),
            upper_uncertainty_col=str(
                ColumnInfo.Name.SOIL_MOISTURE_UNCERTAINTY_VOL_UPPER
            ),
            lower_uncertainty_col=str(
                ColumnInfo.Name.SOIL_MOISTURE_UNCERTAINTY_VOL_LOWER
            ),
            sm_range=sm_range,
            show=self.show_figures,
            backend=self.backend,
            save_location=temp_path,
        )

    @Magazine.reporting_figure(topic="Soil Moisture")
    def _soil_moisture_colour(self):
        """
        Implements colour soil moisture figure.
        """
        temp_path = self.temp_handler.store_figure(
            name="soil_moisture_colour_fig",
            topic=FigureTopic.SOIL_MOISTURE_VOL,
        )

        soil_moisture_coloured_figure(
            data_frame=self.data_frame,
            station_name=self.sensor_info.name,
            sm_column_name=str(ColumnInfo.Name.SOIL_MOISTURE_VOL_FINAL),
            save_location=temp_path,
        )

    @Magazine.reporting_figure(topic="Atmospheric Conditions")
    def _atmospheric_variables(self):
        """
        Implements 3 panel figure with atmopsheric variables
        """
        temp_path = self.temp_handler.store_figure(
            name="atmospheric_variables",
            topic=FigureTopic.ATMOSPHERIC,
        )
        temperature_min = (
            self.data_frame[str(ColumnInfo.Name.AIR_TEMPERATURE)].min() - 2
        )
        temperature_max = (
            self.data_frame[str(ColumnInfo.Name.AIR_TEMPERATURE)].max() + 2
        )

        atmospheric_conditions_figure(
            data_frame=self.data_frame,
            station_name=self.sensor_info.name,
            pressure_col=str(ColumnInfo.Name.AIR_PRESSURE),
            temperature_col=str(ColumnInfo.Name.AIR_TEMPERATURE),
            rel_humidity_col=str(ColumnInfo.Name.AIR_RELATIVE_HUMIDITY),
            temperature_range=(temperature_min, temperature_max),
            show=self.show_figures,
            backend=self.backend,
            save_location=temp_path,
        )

    @Magazine.reporting_figure(topic="Neutron Correction")
    def _neutron_correction_factors(self):
        """
        Implements the correction factor figure
        """
        temp_path = self.temp_handler.store_figure(
            name="correction_factors_figure",
            topic=FigureTopic.NEUTRONS,
        )
        correction_factors_figure(
            data_frame=self.data_frame,
            station_name=self.sensor_info.name,
            pressure_corr_col=str(ColumnInfo.Name.PRESSURE_CORRECTION),
            humidity_corr_col=str(ColumnInfo.Name.HUMIDITY_CORRECTION),
            intensity_corr_col=str(ColumnInfo.Name.INTENSITY_CORRECTION),
            biomass_corr_col=str(
                ColumnInfo.Name.ABOVEGROUND_BIOMASS_CORRECTION
            ),
            show=self.show_figures,
            backend=self.backend,
            save_location=temp_path,
        )

    def _create_intended_figures_list(self):
        """
        Create a list of the figures to be created.

        Returns
        -------
        List
            List of figures to be made
        """
        # Get list of figures to create
        if self.selected_figures:
            figures_to_process = [
                name
                for name, metadata in self._figure_registry.items()
                if name in self.selected_figures
                and metadata.topic not in self.ignore_sections
            ]
        elif self.create_all:
            figures_to_process = [
                name
                for name, metadata in self._figure_registry.items()
                if metadata.topic not in self.ignore_sections
            ]
        else:
            figures_to_process = []

        return figures_to_process

    def create_figures(self):
        """
        Creates the figures
        """
        figures_to_process = self._create_intended_figures_list()

        # Create figures
        for figure_name in figures_to_process:
            try:
                metadata = self._figure_registry[figure_name]
                self._validate_required_columns(metadata=metadata)
                method = self._resolve_method(metadata=metadata)
                method()

            except Exception as e:
                print(f"Could not create figure {figure_name}: {str(e)}")



import pytest
from unittest import mock

from neptoon.io.read.config import _get_config_section
from neptoon.config.configuration_input import ConfigurationManager, BaseConfig


@pytest.fixture
def config_object_simple(monkeypatch):
    configs_dict = {
        "sensor": "sensor",
        "process": "process",
    }
    config = ConfigurationManager()
    config._config = configs_dict
    return config


def test_get_config_section_bad(config_object_simple):
    """
    Raises error when unsupported config is asked for.
    i.e., neither `sensor` or `process`
    """
    with pytest.raises(ValueError):
        tmp_config = config_object_simple
        _ = _get_config_section(tmp_config, wanted_config="name")


def test_get_config_section(monkeypatch):
    """Test config sections is collected"""
    configs_dict = {
        "sensor": "sensor",
        "process": "process",
    }
    config = ConfigurationManager()
    monkeypatch.setattr(config, "_configs", configs_dict)

    config_sensor = _get_config_section(config, wanted_config="sensor")
    config_process = _get_config_section(config, wanted_config="process")
    assert config_process == "process"
    assert config_sensor == "sensor"



import pandas as pd
from pathlib import Path
import pytest
from neptoon.data_prep.timestamp_alignment import (
    TimeStampAligner,
    TimeStampAggregator,
)
from neptoon.utils import recalculate_neutron_uncertainty
from neptoon.columns import ColumnInfo


@pytest.fixture
def correct_df():
    data_path = Path(__file__).parent.parent / "test_data" / "io"
    return pd.read_csv(
        data_path / "unprocessed_df.csv",
        index_col=0,
        parse_dates=True,
    )


def test_create_aligner_object_correct_format(correct_df):
    """
    Test the creation of the object. Test checks for when data is
    formatted correctly.
    """

    ts_align = TimeStampAligner(correct_df)
    assert isinstance(ts_align, TimeStampAligner)


def test_create_aligner_object_wrong_format(correct_df):
    """
    Test the creation of the object. Test checks for when data is not
    formatted correctly (i.e., not datetime index).
    """
    bad_df = correct_df
    bad_df.index = bad_df.index.astype(str)
    with pytest.raises(
        ValueError, match="The DataFrame index must be of datetime type"
    ):
        TimeStampAligner(bad_df)


# def test_align_timestamps():
#     """_summary_"""
#     data_path = Path(__file__).parent / "mock_data"
#     data_before_alignment = pd.read_csv(
#         data_path / "unprocessed_df.csv",
#         index_col=0,
#         parse_dates=True,
#     )
#     data_aligned = pd.read_csv(
#         data_path / "processed_df.csv",
#         index_col=0,
#         parse_dates=True,
#     )
#     tsa = TimeStampAligner(data_before_alignment)
#     tsa.align_timestamps()
#     result_df = tsa.return_dataframe()

#     pd.testing.assert_frame_equal(result_df, data_aligned, check_freq=False)


def test_return_dataframe(correct_df):
    ts_align = TimeStampAligner(correct_df)
    df = ts_align.return_dataframe()
    assert isinstance(df, pd.DataFrame)


#############################
#   TimeStamp Aggregation   #
#############################


@pytest.fixture
def time_stamp_aggregator(correct_df):
    ts_agg = TimeStampAggregator(
        data_frame=correct_df,
        output_resolution="1hour",
        max_na_fraction=0.3,
    )
    return ts_agg


## Correct df has 15 min resolution
def test_timedelta_to_freq_str(time_stamp_aggregator):
    ts_agg = time_stamp_aggregator
    assert ts_agg.avg_temporal_scaling_factor == 4
    assert ts_agg.max_na_int == 1


def test_rescale_uncertainty(correct_df):
    data_frame = correct_df
    data_frame[
        str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_UNCERTAINTY)
    ] = 20
    recalculate_neutron_uncertainty(
        data_frame=data_frame,
        temporal_scaling_factor=4,
    )
    assert (
        data_frame[
            str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_UNCERTAINTY)
        ].iloc[0]
        == 10
    )



from pathlib import Path
import typer

from neptoon.workflow import ProcessWithConfig
from neptoon.config import ConfigurationManager
from neptoon.utils.docker_utils import is_running_in_docker

app = typer.Typer()


@app.callback(invoke_without_command=True)
def main(
    processing_config: str = typer.Option(
        None,
        "--processing",
        "-p",
        help="Path to the processing configuration YAML file",
    ),
    sensor_config: str = typer.Option(
        None,
        "--sensor",
        "-s",
        help="Path to the sensor configuration YAML file",
    ),
):
    """
    Process CRNS data using configuration files.


    Example

    -------

    neptoon -p /path/to/process.yaml -s /path/to/sensor.yaml
    """
    if processing_config and sensor_config:
        typer.secho(
            "Processing the sensor data...", fg=typer.colors.GREEN, bold=True
        )
        process_data(processing_config, sensor_config)
    elif processing_config or sensor_config:
        typer.echo(
            typer.style("Error:", fg=typer.colors.RED, bold=True)
            + " Both processing and station configs are required"
        )
        raise typer.Exit(code=1)
    else:
        typer.echo(
            "Type "
            + typer.style("neptoon --help", fg=typer.colors.CYAN, bold=True)
            + " for help."
        )


def process_data(processing_config: str, sensor_config: str):
    """
    Process the data using the supplied config file locations.
    """
    processing_config_path = Path(processing_config)
    sensor_config_path = Path(sensor_config)

    if not processing_config_path.exists():
        typer.echo(
            f"Error: Processing configuration file not found: {processing_config_path}"
        )
        raise typer.Exit(code=1)

    if not sensor_config_path.exists():
        typer.echo(
            f"Error: Station configuration file not found: {sensor_config_path}"
        )
        raise typer.Exit(code=1)
    if is_running_in_docker():
        config = ConfigurationManager(running_in_docker=True)
    else:
        config = ConfigurationManager(running_in_docker=False)

    try:
        config.load_configuration(file_path=sensor_config_path)
        config.load_configuration(file_path=processing_config_path)

        config_processor = ProcessWithConfig(configuration_object=config)
        config_processor.run_full_process()  # Add verbose into run full process later TODO
        typer.echo("Processing completed successfully.")
    except Exception as e:
        typer.echo(f"Error during processing: {str(e)}")
        import traceback

        traceback.print_exc()
        raise typer.Exit(code=1)


if __name__ == "__main__":
    app()



from pathlib import Path
from neptoon.logging import get_logger
from datetime import timedelta
import re
import pandas as pd
import numpy as np
import pandera.pandas as pa
import datetime

from neptoon.columns import ColumnInfo

core_logger = get_logger()


def validate_and_convert_file_path(
    file_path: str | Path | None,
    base: str | Path | None = None,
) -> Path:
    """
    Ensures that file paths are correctly parsed into pathlib.Path
    objects.

    Parameters
    ----------
    file_path : str | Path | None
        The path to the folder or file.
    base : str | Path | None
        Base to add to file (e..g, a custom base directory)

    Returns
    -------
    pathlib.Path | None
        The file_path as a pathlib.Path object.

    Raises
    ------
    ValueError
        Error if string, pathlib.Path, or None not given.
    """

    if file_path is None:
        return None

    p = Path(file_path)

    if p.is_absolute() and base is not None:
        message = (
            f"Your filepath ({file_path}) is an absolute path",
            " and you've provided a base path to append. If you want to append "
            "basepath your filepath must be relative (maybe remove the first /?)",
        )

        raise AttributeError(message)

    if p.is_absolute():
        return p.resolve()
    if base:
        p = Path(base) / p

    return p.resolve()


def timedelta_to_freq_str(time_delta: datetime.timedelta) -> str:
    """Convert a timedelta to a pandas frequency string."""
    total_seconds = time_delta.total_seconds()

    if total_seconds % (86400) == 0:  # Days (86400 = 24 * 60 * 60)
        return f"{int(total_seconds // 86400)}D"
    elif total_seconds % 3600 == 0:  # Hours
        return f"{int(total_seconds // 3600)}h"
    elif total_seconds % 60 == 0:  # Minutes
        return f"{int(total_seconds // 60)}min"
    else:  # Seconds
        return f"{int(total_seconds)}s"


def validate_timestamp_index(data_frame):
    """
    Checks that the index of the dataframe is timestamp (essential
    for aligning the time stamps and using SaQC)

    Parameters
    ----------
    data_frame : pd.DataFrame
        The data frame imported into the TimeStampAligner

    Raises
    ------
    ValueError
        If the index is not datetime type
    """
    if not pd.api.types.is_datetime64_any_dtype(data_frame.index):
        raise ValueError("The DataFrame index must be of datetime type")


def parse_resolution_to_timedelta(
    resolution_str: str,
):
    """
    Parse a string representation of a time resolution and convert
    it to a timedelta object.

    This method takes a string describing a time resolution (e.g.,
    "30 minutes", "2 hours", "1 day") and converts it into a Python
    timedelta object. It supports minutes, hours, and days as units.

    Parameters
    ----------
    resolution_str : str
        A string representing the time resolution. The format should
        be "<number> <unit>", where <number> is a positive integer
        and <unit> is one of the following: - For minutes: "min",
        "minute", "minutes" - For hours: "hour", "hours", "hr",
        "hrs" - For days: "day", "days" The parsing is
        case-insensitive.

    Returns
    -------
    datetime.timedelta
        A timedelta object representing the parsed time resolution.

    Raises
    ------
    ValueError
        If the resolution string format is invalid or cannot be
        parsed.
    ValueError
        If an unsupported time unit is provided.
    """

    pattern = re.compile(r"(\d+)\s*([a-zA-Z]+)")
    match = pattern.match(resolution_str.strip())

    if not match:
        raise ValueError(f"Invalid resolution format: {resolution_str}")

    value, unit = match.groups()
    value = int(value)

    if unit.lower() in ["min", "mins", "minute", "minutes", "m"]:
        return timedelta(minutes=value)
    elif unit.lower() in ["hour", "hours", "hr", "hrs", "h"]:
        return timedelta(hours=value)
    elif unit.lower() in ["day", "days", "d"]:
        return timedelta(days=value)
    else:
        message = f"Unsupported time unit: {unit}"
        core_logger.error(message)
        raise ValueError(message)


def validate_df(df: pd.DataFrame, schema: pa.DataFrameSchema):
    """
    Validates a df against a pandera.pandas DataFrameSchema

    NOTES:
    Keep it lazy to give info of all df issues

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe to validate
    schema : pa.DataFrameSchema
        Pandera Schema to check against
    """
    return schema.validate(df, lazy=True)


def find_temporal_resolution_seconds(data_frame: pd.DataFrame):

    time_res_seconds = (
        data_frame.index.to_series().dropna().diff().median().total_seconds()
    )
    return time_res_seconds


# def find_median_temporal_resolution_seconds(data_frame: pd.DataFrame):

#     time_res_seconds = (
#         data_frame.index.to_series().dropna().diff().median().total_seconds()
#     )
#     return time_res_seconds


def return_df_with_temporal_resolution_seconds(
    data_frame: pd.DataFrame, column_name=None
):
    data_frame["temp"] = (  # TODO
        data_frame.index.to_series().dropna().diff().total_seconds()
    )
    return data_frame


def is_resolution_greater_than(
    resolution_a: str | datetime.timedelta,
    resolution_b: str | datetime.timedelta,
) -> bool:
    """
    Returns True if resolution_a is greater (coarser) than resolution_b

    Parameters
    ----------
    resolution_a : str | datetime.timedelta
        First resolution to compare
    resolution_b : str | datetime.timedelta
        Second resolution to compare

    Returns
    -------
    bool
        True if resolution_a > resolution_b

    Note
    ----
    If resolution is str then it should be in a form such as "1h",
    "6hour" or "1day". It will be auto converted internally to the func
    """
    if isinstance(resolution_a, str):
        resolution_a = parse_resolution_to_timedelta(resolution_a)
    if isinstance(resolution_b, str):
        resolution_b = parse_resolution_to_timedelta(resolution_b)

    if resolution_a > resolution_b:
        return True
    else:
        return False


def recalculate_neutron_uncertainty(
    data_frame: pd.DataFrame,
    temporal_scaling_factor: int | float,
    uncertainty_col_name: str | None = None,
):
    """
    Adjust the staistical uncertainty of neutrons value based on the
    aggregation.

    Parameters
    ----------
    data_frame : pd.DataFrame
        DataFrame with data
    temporal_scaling_factor : int | float
        The scaling factor to adjust from original resolution to
        revised output resolution
    uncertainty_col_name : str | None, optional
        Name of the col, if None will use the default supplied in
        ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_UNCERTAINTY, by
        default None

    Returns
    -------
    _type_
        _description_
    """
    uncertainty_col_name = (
        str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_UNCERTAINTY)
        if uncertainty_col_name is None
        else uncertainty_col_name
    )

    data_frame[uncertainty_col_name] = data_frame[uncertainty_col_name] * (
        1 / np.sqrt(temporal_scaling_factor)
    )
    return data_frame



"""Generate the code reference pages and navigation."""

from pathlib import Path
import mkdocs_gen_files
import ast

verbose = False


def get_module_functions(module_path):
    """Extract function names from a Python file."""
    if verbose:
        print(f"Trying to parse: {module_path}")
    try:
        with open(module_path, "r", encoding="utf-8") as f:
            content = f.read()

        if not content.strip():
            if verbose:
                print(f"  -> File is empty")
            return [], []

        tree = ast.parse(content)

        functions = []
        classes = []

        for node in ast.walk(tree):
            if isinstance(node, ast.FunctionDef):
                if not node.name.startswith("_"):
                    functions.append(node.name)
                    if verbose:
                        print(f"  -> Found function: {node.name}")
            elif isinstance(node, ast.ClassDef):
                if not node.name.startswith("_"):
                    classes.append(node.name)
                    if verbose:
                        print(f"  -> Found class: {node.name}")

        return functions, classes
    except Exception as e:
        print(f"  -> Error parsing {module_path}: {e}")
        return [], []


nav = mkdocs_gen_files.Nav()

package_name = "neptoon"

for path in sorted(Path(package_name).rglob("*.py")):
    if verbose:
        print(f"\nProcessing: {path}")

    module_path = path.relative_to(package_name).with_suffix("")
    doc_path = path.relative_to(package_name).with_suffix(".md")
    full_doc_path = Path("reference", doc_path)

    parts = list(module_path.parts)
    if verbose:
        print(f"Original parts: {parts}")

    if not parts:
        if verbose:
            print("Skipping: empty parts")
        continue

    if parts[-1] == "__init__":
        if len(parts) == 1:
            parts = ["index"]
            doc_path = Path("index.md")
            full_doc_path = Path("reference", "index.md")
        else:
            parts = parts[:-1]
            doc_path = doc_path.with_name("index.md")
            full_doc_path = full_doc_path.with_name("index.md")
    elif parts[-1] == "__main__":
        if verbose:
            print("Skipping: __main__")
        continue

    if not parts:
        if verbose:
            print("Skipping: parts empty after processing")
        continue

    if verbose:
        print(f"Final parts: {parts}")
        print(f"Doc path: {doc_path}")

    nav[tuple(parts)] = doc_path.as_posix()

    with mkdocs_gen_files.open(full_doc_path, "w") as fd:
        if parts == ["index"]:
            page_path = package_name
            breadcrumps = ""
            title = package_name
        else:
            page_path = ".".join([package_name] + parts)
            breadcrumps = ".".join([package_name] + parts[:-1])
            title = parts[-1]

        if verbose:
            print(f"Writing page for: {title}")

        fd.write(f"{breadcrumps}\n\n")
        fd.write(f"# {title}\n\n")

        # Get functions and classes for TOC
        functions, classes = get_module_functions(path)

        if functions or classes:
            # fd.write("## Contents\n\n")
            if classes:
                fd.write("**Classes:**\n\n")
                for cls in classes:
                    full_cls_path = f"{page_path}.{cls}".lower()
                    fd.write(f"- [{cls}](#{full_cls_path})  \n")
                fd.write("\n")
            if functions:
                fd.write("**Functions:**\n\n")
                for func in functions:
                    full_func_path = f"{page_path}.{func}".lower()
                    fd.write(f"- [{func}](#{full_func_path})  \n")
                fd.write("\n")

        fd.write(f"::: {page_path}\n")
        fd.write("    options:\n")
        fd.write("      heading_level: 3\n")

        if verbose:
            print(f"Successfully wrote page for {title}")

    mkdocs_gen_files.set_edit_path(full_doc_path, path)

with mkdocs_gen_files.open("reference/SUMMARY.md", "w") as nav_file:
    nav_file.writelines(nav.build_literate_nav())

print("Generation complete!")



import pandas as pd
import numpy as np
import copy
from datetime import timedelta
from typing import Literal, List
from dataclasses import dataclass, field
import statistics

# from scipy.optimize import minimize
from neptoon.columns import ColumnInfo
from neptoon.corrections import (
    Schroen2017,
    neutrons_to_grav_soil_moisture_desilets_etal_2010,
    neutrons_to_grav_soil_moisture_koehli_etal_2021,
    find_n0,
)

from neptoon.data_prep.conversions import AbsoluteHumidityCreator


def _create_water_equiv_soc(soil_organic_carbon: float):
    return soil_organic_carbon * 0.555


class CalibrationConfiguration:
    """
    User facing configuration class for calibration steps. Here a user
    can add information required for calibration. Defaults can often
    also be used if following similar data structure and naming as
    neptoon uses as standard.
    """

    def __init__(
        self,
        # Settings
        hours_of_data_around_calib: int = 6,
        converge_accuracy: float = 0.01,
        neutron_conversion_method: Literal[
            "desilets_etal_2010", "koehli_etal_2021"
        ] = "desilets_etal_2010",
        koehli_parameters: Literal[
            "Jan23_uranos",
            "Jan23_mcnpfull",
            "Mar12_atmprof",
            "Mar21_mcnp_drf",
            "Mar21_mcnp_ewin",
            "Mar21_uranos_drf",
            "Mar21_uranos_ewin",
            "Mar22_mcnp_drf_Jan",
            "Mar22_mcnp_ewin_gd",
            "Mar22_uranos_drf_gd",
            "Mar22_uranos_ewin_chi2",
            "Mar22_uranos_drf_h200m",
            "Aug08_mcnp_drf",
            "Aug08_mcnp_ewin",
            "Aug12_uranos_drf",
            "Aug12_uranos_ewin",
            "Aug13_uranos_atmprof",
            "Aug13_uranos_atmprof2",
        ] = "Mar21_mcnp_drf",
        horizontal_weight_method: Literal[
            "schroen_etal_2017", "equal"
        ] = "schroen_etal_2017",
        vertical_weight_method: Literal[
            "schroen_etal_2017", "equal"
        ] = "schroen_etal_2017",
        # Calibration Data Settings
        calib_data_date_time_column_name: str | None = None,
        calib_data_date_time_format: str = "%Y-%m-%d %H:%M",
        sample_depth_column: str | None = None,
        distance_column: str | None = None,
        bulk_density_of_sample_column: str | None = None,
        profile_id_column: str | None = None,
        soil_moisture_gravimetric_column: str | None = None,
        soil_organic_carbon_column: str | None = None,
        lattice_water_column: str | None = None,
        # Time Series Data Settings
        temperature_column: str | None = None,
        relative_humidity_column: str | None = None,
        abs_air_humidity_column: str | None = None,
        neutron_column_name: str | None = None,
        air_pressure_column_name: str | None = None,
        value_avg_lattice_water: float | None = None,
        value_avg_bulk_density: float | None = None,
        value_avg_soil_organic_carbon: float | None = None,
    ):
        """
        User-facing configuration for calibration parameters.

        Parameters
        ----------
        hours_of_data_around_calib : int, optional
            Number of hours of neutron count data to include around the
            datetime stamp for calibration. This window is used to
            gather measurements from sensors during the calibration
            period. Default is 6.
        converge_accuracy : float, optional
            The convergence threshold for when finding n0. Default is
            0.01.
        neutron_conversion_method : {"desilets_etal_2010", "koehli_etal_2021"}, optional
            The conversion method used to translate raw neutron counts
            into soil moisture estimates. Options are
            "desilets_etal_2010" or "koehli_etal_2021". Default is
            "desilets_etal_2010".
        koehli_parameters : str, optional
            Parameter set to use when koehli_etal_2021 method is
            selected. Default is "Mar21_mcnp_drf".
        horizontal_weight_method : {"schroen_etal_2017", "equal"},
        optional
            Method for horizontal weighting. Default is
            "schroen_etal_2017".
        vertical_weight_method : {"schroen_etal_2017", "equal"},
        optional
            Method for vertical weighting. Default is
            "schroen_etal_2017".
        calib_data_date_time_column_name : str, optional
            The name of the column containing date‐time information for
            each calibration day. If None, uses default from ColumnInfo.
        calib_data_date_time_format : str, optional
            Format string for parsing date-time values. Default is
            "%Y-%m-%d %H:%M".
        sample_depth_column : str, optional
            The name of the column with sample depth values (cm). If
            None, uses default from ColumnInfo.
        distance_column : str, optional
            The name of the column stating the distance of the sample
            from the sensor (meters). If None, uses default from
            ColumnInfo.
        bulk_density_of_sample_column : str, optional
            The name of the column with bulk density values of the
            samples (g/cm^3). If None, uses default from ColumnInfo.
        profile_id_column : str, optional
            Name of the column with profile IDs. If None, uses default
            from ColumnInfo.
        soil_moisture_gravimetric_column : str, optional
            Name of the column with gravimetric soil moisture values
            (g/g). If None, uses default from ColumnInfo.
        soil_organic_carbon_column : str, optional
            Name of the column with soil organic carbon values (g/g). If
            None, uses default from ColumnInfo.
        lattice_water_column : str, optional
            Name of the column with lattice water values (g/g). If None,
            uses default from ColumnInfo.
        temperature_column : str, optional
            Name of the column with temperature values. If None, uses
            default from ColumnInfo.
        relative_humidity_column : str, optional
            Name of the column with relative humidity values. If None,
            uses default from ColumnInfo.
        abs_air_humidity_column : str, optional
            Name of the column with absolute air humidity values
            (g/cm3). If None, uses default from ColumnInfo.
        neutron_column_name : str, optional
            Name of the column with corrected neutrons in it. If None,
            uses default from ColumnInfo.
        air_pressure_column_name : str, optional
            Name of the column with air pressure values in it. If None,
            uses default from ColumnInfo.
        value_avg_lattice_water: float, optional
            The actual site average lattice water value. Default is 0.
        value_avg_bulk_density: float, optional
            The actual site average dry soil bulk density. Default is 0.
        value_avg_soil_organic_carbon: float, optional
            The actual site average soil organic carbon. Default is 0.
        """
        # Processing settings
        self.hours_of_data_around_calib = hours_of_data_around_calib
        self.converge_accuracy = converge_accuracy
        self.neutron_conversion_method = neutron_conversion_method
        self.koehli_parameters = koehli_parameters
        self.horizontal_weight_method = horizontal_weight_method
        self.vertical_weight_method = vertical_weight_method

        # Date/time settings
        self.calib_data_date_time_column_name = (
            calib_data_date_time_column_name
        )
        self.calib_data_date_time_format = calib_data_date_time_format

        # Column names (user can override defaults by providing values)
        self.sample_depth_column = sample_depth_column
        self.distance_column = distance_column
        self.bulk_density_of_sample_column = bulk_density_of_sample_column
        self.profile_id_column = profile_id_column
        self.soil_moisture_gravimetric_column = (
            soil_moisture_gravimetric_column
        )
        self.soil_organic_carbon_column = soil_organic_carbon_column
        self.lattice_water_column = lattice_water_column
        self.temperature_column = temperature_column
        self.relative_humidity_column = relative_humidity_column
        self.abs_air_humidity_column = abs_air_humidity_column
        self.neutron_column_name = neutron_column_name
        self.air_pressure_column_name = air_pressure_column_name

        # Extra
        # Site average values
        self.value_avg_lattice_water = value_avg_lattice_water
        self.value_avg_bulk_density = value_avg_bulk_density
        self.value_avg_soil_organic_carbon = value_avg_soil_organic_carbon


@dataclass
class CalibrationContext:
    """
    Internal context for data processing - outputs of each stage are
    written here. Contains resolved column names and processed values.
    """

    # Configuration Values @
    # Processing settings
    hours_of_data_around_calib: int | None = None
    converge_accuracy: float | None = None
    neutron_conversion_method: str | None = None
    koehli_parameters: str | None = None
    horizontal_weight_method: str | None = None
    vertical_weight_method: str | None = None

    # Date/time settings
    calib_data_date_time_column_name: str | None = None
    calib_data_date_time_format: str | None = None

    # Resolved column names
    sample_depth_column: str | None = None
    distance_column: str | None = None
    bulk_density_of_sample_column: str | None = None
    profile_id_column: str | None = None
    soil_moisture_gravimetric_column: str | None = None
    soil_organic_carbon_column: str | None = None
    lattice_water_column: str | None = None
    temperature_column: str | None = None
    relative_humidity_column: str | None = None
    abs_air_humidity_column: str | None = None
    neutron_column_name: str | None = None
    air_pressure_column_name: str | None = None

    # Site average values
    value_avg_lattice_water: float | None = None
    value_avg_bulk_density: float | None = None
    value_avg_soil_organic_carbon: float | None = None
    value_avg_soil_organic_carbon_water_equiv: float | None = None

    # Derived values #
    unique_calibration_days: list = field(default_factory=list)
    list_of_data_frames: list = field(default_factory=list)
    list_of_profiles: list = field(default_factory=list)
    calib_day_df_dict: dict = field(default_factory=dict)
    calib_metrics_dict: dict = field(default_factory=dict)
    calibration_results_by_day: dict = field(default_factory=dict)
    weights_df: pd.DataFrame = field(default_factory=pd.DataFrame)

    @classmethod
    def from_config(cls, config: CalibrationConfiguration):
        """
        Convert configuration to context, resolving all None values to defaults
        from ColumnInfo and calculating derived values.

        Parameters
        ----------
        config : CalibrationConfiguration
            User configuration object

        Returns
        -------
        CalibrationContext
            Internal processing context with all values resolved
        """
        # Resolve column names - use config value if provided, otherwise default
        calib_data_date_time_column_name = (
            config.calib_data_date_time_column_name
            if config.calib_data_date_time_column_name is not None
            else str(ColumnInfo.Name.DATE_TIME)
        )

        sample_depth_column = (
            config.sample_depth_column
            if config.sample_depth_column is not None
            else str(ColumnInfo.Name.CALIB_DEPTH_OF_SAMPLE)
        )

        distance_column = (
            config.distance_column
            if config.distance_column is not None
            else str(ColumnInfo.Name.CALIB_DISTANCE_TO_SENSOR)
        )

        bulk_density_of_sample_column = (
            config.bulk_density_of_sample_column
            if config.bulk_density_of_sample_column is not None
            else str(ColumnInfo.Name.CALIB_BULK_DENSITY)
        )

        profile_id_column = (
            config.profile_id_column
            if config.profile_id_column is not None
            else str(ColumnInfo.Name.CALIB_PROFILE_ID)
        )

        soil_moisture_gravimetric_column = (
            config.soil_moisture_gravimetric_column
            if config.soil_moisture_gravimetric_column is not None
            else str(ColumnInfo.Name.CALIB_SOIL_MOISTURE_GRAVIMETRIC)
        )

        soil_organic_carbon_column = (
            config.soil_organic_carbon_column
            if config.soil_organic_carbon_column is not None
            else str(ColumnInfo.Name.CALIB_SOIL_ORGANIC_CARBON)
        )

        lattice_water_column = (
            config.lattice_water_column
            if config.lattice_water_column is not None
            else str(ColumnInfo.Name.CALIB_LATTICE_WATER)
        )

        temperature_column = (
            config.temperature_column
            if config.temperature_column is not None
            else str(ColumnInfo.Name.AIR_TEMPERATURE)
        )

        relative_humidity_column = (
            config.relative_humidity_column
            if config.relative_humidity_column is not None
            else str(ColumnInfo.Name.AIR_RELATIVE_HUMIDITY)
        )

        abs_air_humidity_column = (
            config.abs_air_humidity_column
            if config.abs_air_humidity_column is not None
            else str(ColumnInfo.Name.ABSOLUTE_HUMIDITY)
        )

        neutron_column_name = (
            config.neutron_column_name
            if config.neutron_column_name is not None
            else str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_FINAL)
        )

        air_pressure_column_name = (
            config.air_pressure_column_name
            if config.air_pressure_column_name is not None
            else str(ColumnInfo.Name.AIR_PRESSURE)
        )

        if config.value_avg_soil_organic_carbon:
            value_avg_soil_organic_carbon_water_equiv = (
                _create_water_equiv_soc(config.value_avg_soil_organic_carbon)
            )
        else:
            value_avg_soil_organic_carbon_water_equiv = None

        return cls(
            # Processing settings
            hours_of_data_around_calib=config.hours_of_data_around_calib,
            converge_accuracy=config.converge_accuracy,
            neutron_conversion_method=config.neutron_conversion_method,
            koehli_parameters=config.koehli_parameters,
            horizontal_weight_method=config.horizontal_weight_method,
            vertical_weight_method=config.vertical_weight_method,
            # Date/time settings
            calib_data_date_time_column_name=calib_data_date_time_column_name,
            calib_data_date_time_format=config.calib_data_date_time_format,
            # Resolved column names
            sample_depth_column=sample_depth_column,
            distance_column=distance_column,
            bulk_density_of_sample_column=bulk_density_of_sample_column,
            profile_id_column=profile_id_column,
            soil_moisture_gravimetric_column=soil_moisture_gravimetric_column,
            soil_organic_carbon_column=soil_organic_carbon_column,
            lattice_water_column=lattice_water_column,
            temperature_column=temperature_column,
            relative_humidity_column=relative_humidity_column,
            abs_air_humidity_column=abs_air_humidity_column,
            neutron_column_name=neutron_column_name,
            air_pressure_column_name=air_pressure_column_name,
            # Site average values
            value_avg_lattice_water=config.value_avg_lattice_water,
            value_avg_bulk_density=config.value_avg_bulk_density,
            value_avg_soil_organic_carbon=config.value_avg_soil_organic_carbon,
            value_avg_soil_organic_carbon_water_equiv=value_avg_soil_organic_carbon_water_equiv,
        )


def _ensure_date_time_index(
    data_frame: pd.DataFrame, context: CalibrationContext
):
    """
    Ensures datetime is properly set as index. If user already has datetime index,
    leave it alone. If datetime is a column, convert to datetime and set as index.

    Parameters
    ----------
    data_frame : pd.DataFrame
        Calibration or time series DataFrame
    context : CalibrationContext
        Context info

    Returns
    -------
    pd.DataFrame
        DataFrame with datetime index

    Raises
    ------
    ValueError
        If the expected datetime column is found neither as column nor index
    """
    target_col = context.calib_data_date_time_column_name
    if (
        (data_frame.index.name == target_col)
        or (
            hasattr(data_frame.index, "names")
            and target_col in data_frame.index.names
        )
        or (
            isinstance(data_frame.index, pd.DatetimeIndex)
            and target_col not in data_frame.columns
        )
    ):
        print(f"Datetime already set as index ('{target_col}'), leaving as-is")
        # Ensure it's datetime type if it isn't already
        if not isinstance(data_frame.index, pd.DatetimeIndex):
            data_frame.index = pd.to_datetime(
                data_frame.index,
                utc=True,
                dayfirst=True,
                format=context.calib_data_date_time_format,
            )
        return data_frame

    elif target_col in data_frame.columns:
        print(f"Converting '{target_col}' column to datetime index")
        data_frame[target_col] = pd.to_datetime(
            data_frame[target_col],
            utc=True,
            dayfirst=True,
            format=context.calib_data_date_time_format,
        )
        data_frame = data_frame.set_index(target_col)
        return data_frame

    else:
        available_cols = list(data_frame.columns)
        index_name = data_frame.index.name

        raise ValueError(
            f"Expected datetime column '{target_col}' not found as column or index. "
            f"Available columns: {available_cols}. "
            f"Index name: {index_name}. "
            f"Please specify the correct column name in CalibrationConfiguration using "
            f"calib_data_date_time_column_name parameter."
        )


class CalibrationStation:
    """
    Abstract which does the complete claibration steps. Can be used on
    its own, but is mainly designed to facilitate CRNSDataHub
    calibration. Simply include the calibration data, the time series
    data and the config object and run find_n0_value(), to return the
    optimum N0.
    """

    def __init__(
        self,
        calibration_data: pd.DataFrame,
        time_series_data: pd.DataFrame,
        config: CalibrationConfiguration,
    ):
        self.calibration_data = calibration_data
        self.time_series_data = (
            time_series_data.copy()
        )  # copy time series to avoid side effects
        self.context = CalibrationContext().from_config(config=config)

    def _collect_stats_for_magazine(self):
        self.number_calib_days = len(
            self.calibrator.return_output_dict_as_dataframe()
        )

    def find_n0_value(self):
        """
        Runs the full process to obtain an N0 estimate.

        Returns
        -------
        float
            N0 estimate after calibration.
        """
        # Prepare calibration data
        calib_prepper = PrepareCalibrationData(
            calibration_data_frame=self.calibration_data,
            context=self.context,
        )
        self.context = calib_prepper.prepare_calibration_data()

        # Prepare neutron data
        times_series_prepper = PrepareNeutronCorrectedData(
            corrected_neutron_data_frame=self.time_series_data,
            context=self.context,
        )
        self.context = times_series_prepper.extract_calibration_day_values()

        # Weight samples of sm
        self.calibrator = CalibrationWeightsCalculator(context=self.context)
        self.context = self.calibrator.calculate_all_sample_weights()

        # Find optimal N0
        n0_finder = CalculateN0(context=self.context)
        optimal_n0 = n0_finder.find_optimal_N0()
        return optimal_n0

    def return_calibration_results_data_frame(self):
        """
        Returns the daily results as a data frame. When multiple days
        calibration is undertaken on each day. The outputs of this are
        saved and this method returns them for viewing.

        Returns
        -------
        pd.DataFrame
            data frame with the results in it.
        """
        return self.calibrator.return_output_dict_as_dataframe()

    def return_weighting_dataframe(self):
        """
        Returns the information about the weighting procedure

        Returns
        -------
        pd.DataFrame
            dataframe with weights
        """
        try:
            return self.context.weights_df
        except Exception as e:
            print(e)


class SampleProfile:

    latest_pid = 0

    def __init__(
        self,
        soil_moisture_gravimetric,
        depth,
        bulk_density,
        site_avg_bulk_density,
        site_avg_organic_carbon,
        site_avg_lattice_water,
        calibration_day,
        distance=1,
        lattice_water=None,
        soil_organic_carbon=None,
        pid=None,
    ):
        """
        Initialise SampleProfile instance.

        Parameters
        ----------
        soil_moisture_gravimetric : array
            array of soil moisture gravimetric values in g/g
        depth : array
            The depth of each soil moisture sample
        bulk_density : array
            bulk density of the samples in g/cm^3
        distance : int, optional
            distance of the profile from the sensor, by default 1
        lattice_water : array-like, optional
            Lattice water from the samples , by default 0
        soil_organic_carbon : int, optional
            _description_, by default 0
        pid : _type_, optional
            _description_, by default None
        """

        # Vector data
        if pid is None:
            SampleProfile.latest_pid += 1
            self.pid = SampleProfile.latest_pid
        else:
            self.pid = pid

        self.soil_moisture_gravimetric = np.array(soil_moisture_gravimetric)
        self.depth = np.array(depth)
        self.bulk_density = np.array(bulk_density)
        self.site_avg_bulk_density = site_avg_bulk_density
        self.calibration_day = calibration_day
        self.soil_organic_carbon = (
            np.array(soil_organic_carbon)
            if soil_organic_carbon is not None
            else np.zeros_like(soil_moisture_gravimetric)
        )
        self.site_avg_organic_carbon = site_avg_organic_carbon
        self.site_avg_organic_carbon_water_equiv = _create_water_equiv_soc(
            site_avg_organic_carbon
        )
        self.lattice_water = self.lattice_water = (
            np.array(lattice_water)
            if lattice_water is not None
            else np.zeros_like(soil_moisture_gravimetric)
        )
        self.site_avg_lattice_water = site_avg_lattice_water
        self.vertical_weights = np.ones_like(soil_moisture_gravimetric)
        self._calculate_sm_total_vol()
        self._calculate_sm_total_grv()

        self._distance = distance
        self.rescaled_distance = distance  # initialise as distance first
        self.sm_total_weighted_avg_grv = np.nan
        self.sm_total_weighted_avg_vol = np.nan
        self.horizontal_weight = 1  # intialise as 1

    @property
    def distance(self):
        return self._distance

    def _calculate_sm_total_vol(self):
        """
        Calculate total volumetric soil moisture.
        """
        # replace nan entries in bulk_density by the site average
        bd = np.where(
            np.isnan(self.bulk_density),
            self.site_avg_bulk_density,
            self.bulk_density,
        )
        self.sm_total_vol = self.soil_moisture_gravimetric * bd

    def _calculate_sm_total_grv(self):
        """
        Calculate total gravimetric soil moisture.
        """

        self.sm_total_grv = self.soil_moisture_gravimetric


class PrepareCalibrationData:
    """
    Prepares the calibration dataframe for processing.

    - ensures datetime index
    - splits multiple days into individual data frames
    - for each calibration day, converts data into a list of
      SampleProfiles
    - calculates site averages of key information (e.g., bulk density)
    - gap fills key info with averages when missing
    """

    def __init__(
        self,
        calibration_data_frame: pd.DataFrame,
        context: CalibrationContext,
    ):
        """
        Instantiate attributes

        Parameters
        ----------
        calibration_data_frame : pd.DataFrame
            The dataframe with the calibration sample data in it. If
            multiple calibration days are available these should be
            stacked in the same dataframe.
        """

        self.calibration_data_frame = calibration_data_frame
        self.context = context

    def _create_list_of_df(self, context: CalibrationContext):
        """
        Splits up the self.calibration_data_frame into individual data
        frames, where each data frame is a different calibration day.

        Parameters
        ----------
        context : CalibrationContext
            Context for calibration

        Returns
        -------
        context
            CalibrationContext
        """

        context.list_of_data_frames = [
            self.calibration_data_frame[
                self.calibration_data_frame.index == calibration_day
            ]
            for calibration_day in context.unique_calibration_days
        ]
        return context

    def _create_calibration_day_profiles(
        self,
        single_day_data_frame,
        context: CalibrationContext,
    ):
        """
        Returns a list of SampleProfile objects which have been created
        from a single calibration day data frame.

        Parameters
        ----------
        single_day_data_frame : pd.DataFrame
            DataFrame snipped during calibration period
        context : CalibrationContext
            Data for processing.

        Returns
        -------
        List of SampleProfiles
            A list of created SampleProfiles
        """
        calibration_day_profiles = []
        profile_ids = np.unique(
            single_day_data_frame[context.profile_id_column]
        )
        for pid in profile_ids:
            temp_df = single_day_data_frame[
                single_day_data_frame[context.profile_id_column] == pid
            ]
            soil_profile = self._create_individual_profile(
                pid=pid, profile_data_frame=temp_df, context=context
            )

            calibration_day_profiles.append(soil_profile)
        return calibration_day_profiles

    def _create_individual_profile(
        self,
        pid,
        profile_data_frame,
        context,
    ):
        """
        Creates a SampleProfile object from a individual profile
        dataframe

        Parameters
        ----------
        pid : numeric
            The profile ID to represent the profile.
        profile_data_frame : pd.DataFrame
            A data frame which holds the values for one single profile.
        context : CalibrationContext
            Information for processing

        Returns
        -------
        SampleProfile
            A SampleProfile object is returned.
        """
        distances = profile_data_frame[context.distance_column].median()
        depths = profile_data_frame[context.sample_depth_column]

        soil_moisture_gravimetric = profile_data_frame[
            context.soil_moisture_gravimetric_column
        ]

        try:
            bulk_density = profile_data_frame[
                context.bulk_density_of_sample_column
            ]
        except KeyError:
            bulk_density = [np.nan] * len(soil_moisture_gravimetric)
        try:
            soil_organic_carbon = profile_data_frame[
                context.soil_organic_carbon_column
            ]
        except KeyError:
            soil_organic_carbon = [np.nan] * len(soil_moisture_gravimetric)

        try:
            lattice_water = profile_data_frame[context.lattice_water_column]
        except KeyError:
            lattice_water = [np.nan] * len(soil_moisture_gravimetric)

        # only need one calibration datetime
        calibration_datetime = next(iter(profile_data_frame.index))

        soil_profile = SampleProfile(
            soil_moisture_gravimetric=soil_moisture_gravimetric,
            depth=depths,
            bulk_density=bulk_density,
            site_avg_bulk_density=context.value_avg_bulk_density,
            distance=distances,
            lattice_water=lattice_water,
            soil_organic_carbon=soil_organic_carbon,
            pid=pid,
            calibration_day=calibration_datetime,
            site_avg_lattice_water=context.value_avg_lattice_water,
            site_avg_organic_carbon=context.value_avg_soil_organic_carbon,
        )
        return soil_profile

    def _create_site_avg_bulk_density(self, context: CalibrationContext):
        """
        Produces a average bulk density from provided sample data, if
        not available uses user provided average, if thats not available
        raises error

        Returns
        -------
        float
            site average bulk density

        Raises
        ------
        ValueError
            No bulk density provided
        """
        if (
            context.bulk_density_of_sample_column
            not in self.calibration_data_frame.columns
        ):
            message = (
                "There appears to be no bulk density data in the calibration sample dataset. "
                "Attempting to use the provided site average..."
            )
            print(message)
            if context.value_avg_bulk_density is None:
                message = (
                    "There is no provided site average bulk density value. Please provide this "
                    "in sensor configuration files (if using configs), SensorInformation (if using notebooks)"
                    "or CalibrationConfiguration (if using calibration module directly.)"
                )
                raise ValueError(message)
            else:
                message = f"Using the provided site average bulk density value: {context.value_avg_bulk_density}"
                print(message)

        else:
            message = "Calculating site average bulk density from provided sample data."
            print(message)
            context.value_avg_bulk_density = self.calibration_data_frame[
                context.bulk_density_of_sample_column
            ].mean()

        return context

    def _create_site_avg_lattice_water(self, context: CalibrationContext):
        """
        Produces a average lattice water from provided sample data, if
        not available uses user provided average, if thats not available
        raises error. Adds info to context

        Returns
        -------
        context
            Context

        Raises
        ------
        ValueError
            No lattice water provided
        """
        if (
            context.lattice_water_column
            not in self.calibration_data_frame.columns
        ):
            message = (
                "There appears to be no lattice water data in the calibration sample dataset. "
                "Attempting to use the provided site average..."
            )
            print(message)
            if context.value_avg_lattice_water is None:
                message = (
                    "There is no provided site average lattice water value. Please provide this "
                    "in sensor configuration files (if using configs), SensorInformation (if using notebooks)"
                    "or CalibrationConfiguration (if using calibration module directly.)\n"
                    "NOTE: If you don't know, use 0"
                )
                raise ValueError(message)
            else:
                message = f"Using the provided site average lattice water value: {context.value_avg_lattice_water}."
                print(message)

        else:
            message = "Calculating site average lattice water from provided sample data."
            print(message)
            context.value_avg_lattice_water = self.calibration_data_frame[
                context.lattice_water_column
            ].mean()
        return context

    def _create_site_avg_soil_organic_carbon(
        self, context: CalibrationContext
    ):
        """
        Produces a average soil_organic_carbon from provided sample
        data, if not available uses user provided average, if thats not
        available raises error

        Returns
        -------
        CalibrationContext
            site average soil_organic_carbon in context

        Raises
        ------
        ValueError
            No soil_organic_carbon provided
        """
        if (
            context.soil_organic_carbon_column
            not in self.calibration_data_frame.columns
        ):
            message = (
                "There appears to be no soil_organic_carbon data in the calibration sample dataset. "
                "Attempting to use the provided site average..."
            )
            print(message)
            if context.value_avg_soil_organic_carbon is None:
                message = (
                    "There is no provided site average soil_organic_carbon value. Please provide this "
                    "in sensor configuration files (if using configs), SensorInformation (if using notebooks)"
                    "or CalibrationConfiguration (if using calibration module directly.)\n"
                    "NOTE: If you don't know, use 0"
                )
                raise ValueError(message)
            else:
                message = f"Using the provided site average soil_organic_carbon value: {context.value_avg_soil_organic_carbon}"
                print(message)

        else:
            message = "Calculating site average soil_organic_carbon from provided sample data."
            print(message)

            context.value_avg_soil_organic_carbon = (
                self.calibration_data_frame[
                    context.soil_organic_carbon_column
                ].mean()
            )

        if np.isnan(context.value_avg_soil_organic_carbon):
            context.value_avg_soil_organic_carbon = 0

        context.value_avg_soil_organic_carbon_water_equiv = (
            _create_water_equiv_soc(context.value_avg_soil_organic_carbon)
        )

        return context

    def _create_site_avg_values(self, context: CalibrationContext):
        """
        Derives site avg values required for calibration

        Returns
        -------
        CalibrationContext
            more context
        """
        context = self._create_site_avg_bulk_density(context=context)
        context = self._create_site_avg_lattice_water(context=context)
        context = self._create_site_avg_soil_organic_carbon(context=context)
        return context

    def _parse_unique_calibration_days(
        self,
        context: CalibrationContext,
    ):
        """
        Parses unique calibration days and adds information to context

        Parameters
        ----------
        context : CalibrationContext
            DataContext for calibration

        Returns
        -------
        context
            CalibrationContext
        """
        context.unique_calibration_days = np.unique(
            self.calibration_data_frame.index
        )
        return context

    def _process_dfs_into_profiles(self, context: CalibrationContext):
        """
        Processes the calibration day data into profiles

        Parameters
        ----------
        context : CalibrationContext
            _description_

        Returns
        -------
        CalibrationContext
            context with profiles
        """

        for data_frame in context.list_of_data_frames:
            calibration_day_profiles = self._create_calibration_day_profiles(
                single_day_data_frame=data_frame,
                context=context,
            )
            context.list_of_profiles.extend(calibration_day_profiles)
        return context

    def prepare_calibration_data(self):
        """
        Prepares the calibration data into a list of profiles.
        """
        context = self.context
        self.calibration_data_frame = _ensure_date_time_index(
            data_frame=self.calibration_data_frame, context=context
        )

        context = self._parse_unique_calibration_days(context)
        context = self._create_site_avg_values(context)
        context = self._create_list_of_df(context=context)
        context = self._process_dfs_into_profiles(context=context)

        return context


class PrepareNeutronCorrectedData:

    def __init__(
        self,
        corrected_neutron_data_frame: pd.DataFrame,
        context: CalibrationConfiguration,
    ):
        self.corrected_neutron_data_frame = corrected_neutron_data_frame
        self.context = context

    def _ensure_abs_humidity_exists(self, data_frame: pd.DataFrame):
        """
        Checks to see if absolute humidity exists in the data frame. If
        it doesn't it will create it.

        Parameters
        ----------
        data_frame : pd.DataFrame
            Corrected neutron dataframe

        Returns
        -------
        pd.DataFrame
            Corrected neutron dataframe with abs humidity
        """
        if str(ColumnInfo.Name.ABSOLUTE_HUMIDITY) not in data_frame.columns:
            abs_humidity_creator = AbsoluteHumidityCreator(
                data_frame=data_frame
            )
            data_frame = abs_humidity_creator.check_and_return_abs_hum_column()
            return data_frame
        else:
            return data_frame

    def _uncorrect_humidity_for_koehli(
        self,
        data_frame: pd.DataFrame,
        context: CalibrationContext,
    ):
        """
        Uncorrect humidity from corrected neutrons if this was done and
        the koehli method is selected.

        NOTE: This occurs on a copy of the crns_data_frame (i.e., the
        changes are not maintained outside of the calibration stage). It
        will be uncorrected a second time later when the conversion to
        soil moisture happens, if required.

        Parameters
        ----------
        data_frame : pd.DataFrame
            DataFrame with data for processing
        context : _type_, optional
            context, by default CalibrationContext

        Returns
        -------
        pd.DataFrame
            DataFrame with either correction removed or not
        """
        if (
            context.neutron_conversion_method == "koehli_etal_2021"
            and str(ColumnInfo.Name.HUMIDITY_CORRECTION) in data_frame.columns
        ):
            data_frame[context.neutron_column_name] /= data_frame[
                str(ColumnInfo.Name.HUMIDITY_CORRECTION)
            ]
            return data_frame
        else:
            return data_frame

    def extract_calibration_day_values(self):
        """
        Extracts the rows of data for each calibration day.
        """

        context = self.context
        data_frame = self.corrected_neutron_data_frame
        data_frame = _ensure_date_time_index(
            data_frame=data_frame, context=context
        )
        data_frame = self._ensure_abs_humidity_exists(data_frame=data_frame)
        data_frame = self._uncorrect_humidity_for_koehli(
            data_frame=data_frame, context=context
        )

        calibration_indicies_dict = self._extract_calibration_day_indices(
            corrected_neutron_data_frame=data_frame, context=context
        )

        dict_of_data = {}
        for value in calibration_indicies_dict.values():
            tmp_df = data_frame.loc[value]
            calib_day = None
            # Find calibration day index to use as dict key
            for day in context.unique_calibration_days:
                calib_day = self._find_nearest_calib_day_in_indicies(
                    day=day, data_frame=tmp_df
                )
                if calib_day is not None:
                    break
            dict_of_data[calib_day] = tmp_df

        context.calib_day_df_dict = dict_of_data
        return context

    def _find_nearest_calib_day_in_indicies(self, day, data_frame):
        """
        Finds the nearest calibration day to the indices.

        Parameters
        ----------
        day : Timedelta
            _description_
        data_frame : pd.DataFrame
            dataframe of data

        Returns
        -------
        Timedelta
            Calibration day
        """

        day = pd.to_datetime(day)
        mask = (data_frame.index >= day - timedelta(hours=1)) & (
            data_frame.index <= day + timedelta(hours=1)
        )
        if mask.any():
            calib_day = day
            return calib_day

    def _extract_calibration_day_indices(
        self,
        corrected_neutron_data_frame,
        context: CalibrationContext,
    ):
        """
        Extracts the required indices

        Parameters
        ----------
        corrected_neutron_data_frame : pd.DataFrame
            DataFrame with corrected neutrons and meteorological data
        context : CalibrationContext
            Context for processing

        Returns
        -------
        dict
            A dictionary for each calibration date with the indices to
            extract from corrected neutron data.
        """
        extractor = IndicesExtractor(
            corrected_neutron_data_frame=corrected_neutron_data_frame,
            context=context,
        )
        calibration_indices = extractor.extract_calibration_day_indices()

        return calibration_indices


class IndicesExtractor:
    """
    Extracts indices from the corrected neutron data based on the
    supplied calibration days
    """

    def __init__(
        self,
        corrected_neutron_data_frame,
        context,
    ):
        """
        Attributes.

        Parameters
        ----------
        corrected_neutron_data_frame : pd.DataFrame
            The corrected neutron data frame
        calibration_data_prepper : PrepareCalibrationData
            The processed object
        hours_of_data_to_extract : int, optional
            The number of hours of data around the calibration date time
            stamp to collect., by default 6
        """
        self.corrected_neutron_data_frame = corrected_neutron_data_frame
        self.context = context

    def _convert_to_datetime(
        self,
        dates,
    ):
        """
        Convert a list of dates to pandas Timestamp objects.
        """
        return pd.to_datetime(dates)

    def _create_time_window(
        self,
        date: pd.Timestamp,
        context: CalibrationContext,
    ):
        """
        Create a time window around a given date.

        Parameters
        ----------
        date : pd.Timestamp
            Time stamp to create windows around
        context : CalibrationContext
            Context for processing

        Returns
        -------
        Timedelta, Timedelta
            Start and end Timedelta for calib window
        """
        half_window = context.hours_of_data_around_calib / 2
        window = pd.Timedelta(hours=half_window)
        return date - window, date + window

    def _extract_indices_within_window(
        self,
        start: pd.Timestamp,
        end: pd.Timestamp,
        data_frame: pd.DataFrame,
    ):
        """
        Extract indices of data points within a given time window.

        Parameters
        ----------
        start : pd.Timestamp
            Start point
        end : pd.Timestamp
            End point
        data_frame : pd.DataFrame
            DataFrame

        Returns
        -------
        List
            Indexes in range
        """
        mask = (data_frame.index >= start) & (data_frame.index <= end)
        return data_frame.index[mask].tolist()

    def extract_calibration_day_indices(self):
        """
        Extract indices for each calibration day within a time window.

        Returns
        -------
        Dict
            Indicies for each calibration day
        """
        context = self.context
        unique_days = self._convert_to_datetime(
            context.unique_calibration_days
        )

        calibration_indices = {}
        for day in unique_days:
            start, end = self._create_time_window(
                date=day,
                context=context,
            )
            calibration_indices[day] = self._extract_indices_within_window(
                start=start,
                end=end,
                data_frame=self.corrected_neutron_data_frame,
            )

        return calibration_indices


class CalibrationWeightsCalculator:
    def __init__(
        self,
        context: CalibrationContext | None = None,
    ):
        self.context = context

    def set_values(
        self,
        config: CalibrationConfiguration,
        calibration_df: pd.DataFrame,
        absolute_humidity: list,
        air_pressure: list,
    ):
        """
        Set required values for calculating weights on calibration data.
        Use this when you already have derived values (e.g., air
        pressure during calibration period), and want to submit them
        directly.

        Parameters
        ----------
        config : CalibrationConfiguration
            Calibration config with required information for finding
            sample weights.
        calibration_df : pd.DataFrame
            Calibration sample data
        absolute_humidity : list
            List of absolute humidity values, one for each calibration
            day in the order they appear in the dataframe.
        air_pressure : list
            List of atmospheric pressure values, one for each
            calibration day in the order they appear in the dataframe.
        bulk_density : float | None, optional
            Bulk Density for the site, by default None
        lattice_water : float | None, optional
            Lattice water, by default None
        soil_organic_carbon : float | None, optional
            Soil organic carbon, by default None
        """

        context = CalibrationContext().from_config(config=config)
        average_neutron_count = np.ones_like(
            absolute_humidity
        )  # Need this to not break code

        calib_df_prepper = PrepareCalibrationData(
            calibration_data_frame=calibration_df, context=context
        )
        context = calib_df_prepper.prepare_calibration_data()

        for i in range(len(absolute_humidity)):
            day = context.unique_calibration_days[i]
            data = {
                context.abs_air_humidity_column: absolute_humidity[i],
                context.air_pressure_column_name: air_pressure[i],
                context.neutron_column_name: average_neutron_count[i],
            }
            tmp_df = pd.DataFrame(data=data, index=[day])
            context.calib_day_df_dict[day] = tmp_df
        self.context = context

    def _get_time_series_data_for_day(self, day):
        """
        Get data for particular day

        Parameters
        ----------
        day : Timedelta
            Day

        Returns
        -------
        _type_
            _description_
        """
        return self.context.calib_day_df_dict[day]

    @staticmethod
    def _initial_vol_sm_estimate(profiles: List):
        """
        Gets an initial equal average soil moisture estimate

        Parameters
        ----------
        profiles : List
            List of SampleProfiles

        Returns
        -------
        sm_estimate : float
            Estimate of field soil moisture
        """
        sm_total_vol_values = [
            np.array(profile.sm_total_vol).flatten()
            for profile in profiles
            if profile.sm_total_vol is not None
        ]
        flattened = np.concatenate(sm_total_vol_values)
        valid_values = flattened[~np.isnan(flattened)]
        sm_estimate = np.mean(valid_values)
        return sm_estimate

    def calculate_all_sample_weights(self):
        """
        Applies the weighting procedure to multiple calibration days if
        present.

        Returns
        -------
        CalibrationContext
            With more context
        """

        context = self.context

        for day in context.unique_calibration_days:

            tmp_data = self._get_time_series_data_for_day(day)
            profiles = [
                p for p in context.list_of_profiles if p.calibration_day == day
            ]

            volumetric_sm_estimate = self._initial_vol_sm_estimate(
                profiles=profiles
            )
            average_abs_air_humidity = tmp_data[
                context.abs_air_humidity_column
            ].mean()
            average_air_pressure = tmp_data[
                context.air_pressure_column_name
            ].mean()
            average_neutron_count = tmp_data[
                context.neutron_column_name
            ].mean()

            field_average_sm_vol, field_average_sm_grav, footprint = (
                self._calculate_weighted_sm_average(
                    profiles=profiles,
                    initial_volumetric_sm_estimate=volumetric_sm_estimate,
                    average_abs_air_humidity=average_abs_air_humidity,
                    average_air_pressure=average_air_pressure,
                )
            )

            output = {
                "field_average_soil_moisture_volumetric": field_average_sm_vol,
                "field_average_soil_moisture_gravimetric": field_average_sm_grav,
                "average_neutron_count": average_neutron_count,
                "horizontal_footprint_radius_in_meters": footprint,
                "absolute_air_humidity": average_abs_air_humidity,
                "atmospheric_pressure": average_air_pressure,
            }

            context.calib_metrics_dict[day] = output

        context = self._sample_profiles_to_dataframe(context)
        return context

    def _calculate_weighted_sm_average(
        self,
        profiles: List,
        initial_volumetric_sm_estimate: float,
        average_abs_air_humidity: float,
        average_air_pressure: float,
    ):
        """
        Calculates the field average soil moisture value according to
        Schrön et al., 2017. Weighting is calculated on volumetric soil
        moisture volumes (as described in the paper), but a weighted
        gravimetric value is given for calibration in neptoon.

        Parameters
        ----------
        profiles : List[Profile]
            A list of soil‐profile objects collected on the same day.
            Each Profile must have:
              - `.rescaled_distance` (rescaled distance from sensor, in
                m)
              - `.site_avg_bulk_density` (bulk density, in g/cm3)
              - `.depth` (array of depths, in cm)
              - `.sm_total_vol` (array of volumetric‐moisture values,
                cm3/cm3)
              - `.sm_total_grv` (array of gravimetric‐moisture values,
                g/g)
        initial_sm_estimate : float
            Initial soil moisture estimate (usually equal average)
        average_abs_air_humidity : float
            Average absolute air humidity
        average_air_pressure : float
            Air pressure average during calibration period (hPa)

        Returns
        -------
        field_average_sm_volumetric : float
            Converged volumetric soil moisture (cm3/cm3).
        field_average_sm_gravimetric : float
            Corresponding converged gravimetric soil moisture (g/g).
        footprint_m : float
            Estimated radius (m) of the footprint of the sensor.

        Notes
        -----
        - Convergence is checked via `abs((new_volumetric_estimate -
          old_estimate) / old_estimate) <
          self.config.converge_accuracy`.
        - Uses `Schroen2017.rescale_distance`,
          `Schroen2017.calculate_measurement_depth`,
          `Schroen2017.vertical_weighting`,
          `Schroen2017.horizontal_weighting`, and
          `Schroen2017.calculate_footprint_radius` at each iteration.

        """

        volumetric_sm_estimate = copy.deepcopy(initial_volumetric_sm_estimate)
        accuracy = 1
        field_average_sm_volumetric = None
        field_average_sm_gravimetric = None

        while accuracy > self.context.converge_accuracy:
            profile_sm_averages_volumetric = []
            profile_sm_averages_gravimetric = []
            profiles_horizontal_weights = []

            for p in profiles:

                p.rescaled_distance = Schroen2017.rescale_distance(
                    distance_from_sensor=p.rescaled_distance,
                    atmospheric_pressure=average_air_pressure,
                    volumetric_soil_moisture=volumetric_sm_estimate,
                )

                if self.context.vertical_weight_method == "equal":
                    p.vertical_weights = np.ones_like(
                        p.soil_moisture_gravimetric
                    )
                else:
                    p.vertical_weights = Schroen2017.vertical_weighting(
                        depth=p.depth,
                        distance=p.rescaled_distance,
                        bulk_density=p.site_avg_bulk_density,
                        volumetric_soil_moisture=volumetric_sm_estimate,
                    )
                    # Normalise
                    p.vertical_weights = p.vertical_weights / sum(
                        p.vertical_weights
                    )

                # Calculate weighted sm average
                p.sm_total_weighted_avg_vol = np.average(
                    p.sm_total_vol, weights=p.vertical_weights
                )
                p.sm_total_weighted_avg_grv = np.average(
                    p.sm_total_grv, weights=p.vertical_weights
                )
                if self.context.horizontal_weight_method == "equal":
                    p.horizontal_weight = 1
                else:
                    p.horizontal_weight = Schroen2017.horizontal_weighting(
                        distance=p.rescaled_distance,
                        volumetric_soil_moisture=p.sm_total_weighted_avg_vol,
                        abs_air_humidity=average_abs_air_humidity,
                    )

                # create a list of average sm and horizontal weights
                profile_sm_averages_volumetric.append(
                    p.sm_total_weighted_avg_vol
                )
                profile_sm_averages_gravimetric.append(
                    p.sm_total_weighted_avg_grv
                )
                profiles_horizontal_weights.append(p.horizontal_weight)

            # mask out nan values from list
            profile_sm_averages_volumetric = np.ma.MaskedArray(
                profile_sm_averages_volumetric,
                mask=np.isnan(profile_sm_averages_volumetric),
            )
            profile_sm_averages_gravimetric = np.ma.MaskedArray(
                profile_sm_averages_gravimetric,
                mask=np.isnan(profile_sm_averages_gravimetric),
            )
            profiles_horizontal_weights = np.ma.MaskedArray(
                profiles_horizontal_weights,
                mask=np.isnan(profiles_horizontal_weights),
            )

            # Normalise horizontal weights
            weight_sum = np.ma.sum(profiles_horizontal_weights)
            profiles_horizontal_weights = (
                profiles_horizontal_weights / weight_sum
            )
            for i, p in enumerate(profiles):
                p.horizontal_weight = profiles_horizontal_weights.data[i]

            # create field averages of soil moisture

            field_average_sm_volumetric = np.average(
                profile_sm_averages_volumetric,
                weights=profiles_horizontal_weights,
            )
            field_average_sm_gravimetric = np.average(
                profile_sm_averages_gravimetric,
                weights=profiles_horizontal_weights,
            )

            # check convergence accuracy
            if (
                self.context.vertical_weight_method == "schroen_etal_2017"
                and self.context.horizontal_weight_method
                == "schroen_etal_2017"
            ):
                accuracy = abs(
                    (field_average_sm_volumetric - volumetric_sm_estimate)
                    / volumetric_sm_estimate
                )
            else:
                # Stop convergence if any weighting left as equal
                accuracy = self.context.converge_accuracy

            if accuracy > self.context.converge_accuracy:

                volumetric_sm_estimate = copy.deepcopy(
                    field_average_sm_volumetric
                )
                profile_sm_averages_volumetric = []
                profile_sm_averages_gravimetric = []
                profiles_horizontal_weights = []

        footprint_m = Schroen2017.calculate_footprint_radius(
            volumetric_soil_moisture=field_average_sm_volumetric,
            abs_air_humidity=average_abs_air_humidity,
            atmospheric_pressure=average_air_pressure,
        )

        return (
            field_average_sm_volumetric,
            field_average_sm_gravimetric,
            footprint_m,
        )

    def return_output_dict_as_dataframe(self):
        """
        Returns the dictionary of information created for each
        calibration day during processing as a pandas DataFrame

        Returns
        -------
        pd.DataFrame
            DataFrame with information created during processing.
        """
        df = pd.DataFrame.from_dict(
            self.context.calib_metrics_dict, orient="index"
        )
        df = df.reset_index()
        df = df.rename(
            columns={
                "index": "calibration_day",
                "field_average_soil_moisture_volumetric": "field_average_soil_moisture_volumetric",
                "field_average_soil_moisture_gravimetric": "field_average_soil_moisture_gravimetric",
                "horizontal_footprint_in_meters": "horizontal_footprint_radius",
            }
        )
        return df

    def return_weighting_dataframe(self):
        """
        Returns the information about the weighting procedure

        Returns
        -------
        pd.DataFrame
            dataframe with weights
        """
        return self.context.weights_df

    def _get_array_attributes(self, profile) -> List[str]:
        """Get list of attributes that are arrays/lists."""
        array_attrs = []

        for attr_name in dir(profile):
            if attr_name.startswith("_") or callable(
                getattr(profile, attr_name)
            ):
                continue

            attr_value = getattr(profile, attr_name)
            if (
                isinstance(attr_value, (np.ndarray, list))
                and len(attr_value) > 1
            ):
                array_attrs.append(attr_name)

        return array_attrs

    def _get_scalar_attributes(self, profile) -> List[str]:
        """Get list of attributes that are scalar values."""
        scalar_attrs = []

        for attr_name in dir(profile):
            if attr_name.startswith("_") or callable(
                getattr(profile, attr_name)
            ):
                continue

            attr_value = getattr(profile, attr_name)
            if not isinstance(attr_value, (np.ndarray, list)) or (
                isinstance(attr_value, (np.ndarray, list))
                and len(attr_value) <= 1
            ):
                # Handle scalar numpy arrays
                if isinstance(attr_value, np.ndarray) and attr_value.size == 1:
                    scalar_attrs.append(attr_name)
                elif not isinstance(attr_value, (np.ndarray, list)):
                    scalar_attrs.append(attr_name)

        return scalar_attrs

    def _sample_profiles_to_dataframe(
        self,
        context: CalibrationContext,
    ):
        """
        Convert a list of SampleProfile objects to a pandas DataFrame.

        Creates a long-format DataFrame where each row represents one
        depth measurement. Scalar values from each profile are repeated
        for all depth measurements in that profile.

        Parameters
        ----------
        context: CalibrationContext
            Context which holds list of profiles

        Returns
        -------
        CalibrationContext
        """
        sample_profiles = context.list_of_profiles
        if not sample_profiles:
            return pd.DataFrame()

        all_rows = []

        for profile in sample_profiles:
            array_attrs = self._get_array_attributes(profile)
            scalar_attrs = self._get_scalar_attributes(profile)

            if not array_attrs:
                row_data = {
                    attr: getattr(profile, attr) for attr in scalar_attrs
                }
                all_rows.append(row_data)
            else:
                array_length = len(getattr(profile, array_attrs[0]))

                for i in range(array_length):
                    row_data = {}

                    for attr in scalar_attrs:
                        row_data[attr] = getattr(profile, attr)

                    for attr in array_attrs:
                        array_val = getattr(profile, attr)
                        row_data[attr] = (
                            array_val[i]
                            if isinstance(array_val, np.ndarray)
                            else array_val[i]
                        )

                    all_rows.append(row_data)

        context.weights_df = pd.DataFrame(all_rows)
        return context


class CalculateN0:
    def __init__(
        self,
        context: CalibrationContext | None = None,
    ):
        self.context = context
        self._using_custom_vals = False
        self._neutron_counts = None  # only used for custom inputs

    def set_values(
        self,
        soil_moisture: list,
        corrected_neutron_counts: list,
        conversion_method: Literal[
            "desilets_etal_2010", "koehli_etal_2021"
        ] = "desilets_etal_2010",
        lattice_water=0,
        water_equiv_soil_organic_carbon=0,
        absolute_humidity=None,
        koehli_parameters: Literal[
            "Jan23_uranos",
            "Jan23_mcnpfull",
            "Mar12_atmprof",
            "Mar21_mcnp_drf",
            "Mar21_mcnp_ewin",
            "Mar21_uranos_drf",
            "Mar21_uranos_ewin",
            "Mar22_mcnp_drf_Jan",
            "Mar22_mcnp_ewin_gd",
            "Mar22_uranos_drf_gd",
            "Mar22_uranos_ewin_chi2",
            "Mar22_uranos_drf_h200m",
            "Aug08_mcnp_drf",
            "Aug08_mcnp_ewin",
            "Aug12_uranos_drf",
            "Aug12_uranos_ewin",
            "Aug13_uranos_atmprof",
            "Aug13_uranos_atmprof2",
        ] = "Mar21_mcnp_drf",
    ):
        def _ensure_list(value, name):
            if value is None:
                return None
            if isinstance(value, (int, float)):
                return [float(value)]
            if isinstance(value, list):
                return [float(x) for x in value]
            raise TypeError(
                f"{name} must be a number, list of numbers, or None"
            )

        soil_moisture = _ensure_list(soil_moisture, "soil_moisture")
        corrected_neutron_counts = _ensure_list(
            corrected_neutron_counts, "neutron_counts"
        )
        self._neutron_counts = corrected_neutron_counts
        if len(soil_moisture) != len(corrected_neutron_counts):
            raise ValueError(
                f"soil_moisture and neutron_count must have the same length. "
                f"Got {len(soil_moisture)} and {len(corrected_neutron_counts)} respectively."
            )

        if absolute_humidity is None:
            absolute_humidity = [0.0] * len(soil_moisture)
        else:
            absolute_humidity = _ensure_list(
                absolute_humidity, "absolute_humidity"
            )
            if len(absolute_humidity) != len(corrected_neutron_counts):
                raise ValueError(
                    f"absolute_humidity and neutron_count must have the same length. "
                    f"Got {len(soil_moisture)} and {len(corrected_neutron_counts)} respectively."
                )

        metrics_dict = {}
        for i in range(len(soil_moisture)):
            output = {
                "field_average_soil_moisture_gravimetric": soil_moisture[i],
                "average_neutron_count": corrected_neutron_counts[i],
                "absolute_air_humidity": absolute_humidity[i],
            }
            metrics_dict[i] = output

        self.context = CalibrationContext(
            neutron_conversion_method=conversion_method,
            value_avg_lattice_water=lattice_water,
            value_avg_soil_organic_carbon_water_equiv=water_equiv_soil_organic_carbon,
            koehli_parameters=koehli_parameters,
            calib_metrics_dict=metrics_dict,
        )
        self._using_custom_vals = True

    def _find_optimal_n0_single_day_desilets_etal_2010(
        self,
        gravimetric_sm_on_day,
        neutron_mean,
        n0_range: pd.Series,
        lattice_water=0,
        water_equiv_soil_organic_carbon=0,
    ):
        """
        Finds optimal N0 number when using desilets et al., 2010 method

        Parameters
        ----------
        gravimetric_sm_on_day : float
            gravimetric soil moisture (weighted)
        neutron_mean : float
            average (corrected) neutron count

        Returns
        -------
        float
            N0
        """
        gravimetric_sm_on_day_total = gravimetric_sm_on_day
        # gravimetric_sm_on_day_total = (
        #     gravimetric_sm_on_day
        #     + lattice_water
        #     + water_equiv_soil_organic_carbon
        # )

        def calculate_sm_and_error_desilets(n0):
            sm_prediction = neutrons_to_grav_soil_moisture_desilets_etal_2010(
                neutron_count=neutron_mean,
                n0=n0,
            )
            rel_error = (
                abs(sm_prediction - gravimetric_sm_on_day_total)
            ) / gravimetric_sm_on_day_total

            return pd.Series(
                {
                    "N0": n0,
                    "soil_moisture_prediction": sm_prediction,
                    "relative_error": rel_error,
                }
            )

        results_df = n0_range.apply(calculate_sm_and_error_desilets)
        return results_df

    def _find_optimal_n0_single_day_koehli_etal_2021(
        self,
        gravimetric_sm_on_day: float,
        neutron_mean: float,
        n0_range: pd.Series,
        abs_air_humidity: float,
        lattice_water: float,
        water_equiv_soil_organic_carbon: float,
        koehli_parameters: str,
    ):
        """
        Finds optimal N0 number when using Koehli etal method

        Parameters
        ----------
        gravimetric_sm_on_day : float
            Average gravimetic water on calibration day)
        neutron_mean : float | int
            Mean corrected neutron count
        abs_air_humidity : float
            Absolute air humidity
        lattice_water : float
            Lattice water content of soil
        water_equiv_soil_organic_carbon : float
            water equivelant of soil organic carbon
        koehli_parameters: str
            The specific method form of Koehli method

        Returns
        -------
        Tuple
            The N0 calibration term and absolute error (dummy nan value)
        """
        # accounted for already in functions
        gravimetric_sm_on_day_total = gravimetric_sm_on_day
        # gravimetric_sm_on_day_total = (
        #     gravimetric_sm_on_day
        #     + lattice_water
        #     + water_equiv_soil_organic_carbon
        # )

        def calculate_sm_and_error_koehli(n0):

            sm_prediction = neutrons_to_grav_soil_moisture_koehli_etal_2021(
                neutron_count=neutron_mean,
                n0=n0,
                abs_air_humidity=abs_air_humidity,
                additional_gravimetric_water=lattice_water
                + water_equiv_soil_organic_carbon,
                # lattice_water=lattice_water,
                # water_equiv_soil_organic_carbon=water_equiv_soil_organic_carbon,
                koehli_parameters=koehli_parameters,
            )
            rel_error = (
                abs(sm_prediction - gravimetric_sm_on_day_total)
            ) / gravimetric_sm_on_day_total

            return pd.Series(
                {
                    "N0": n0,
                    "soil_moisture_prediction": sm_prediction,
                    "relative_error": rel_error,
                }
            )

        results_df = n0_range.apply(calculate_sm_and_error_koehli)
        return results_df

    def _create_n0_range(
        self,
        context: CalibrationContext,
        custom_range=False,
        neutron_counts=None,
    ):
        """
        Create a range of n0

        Parameters
        ----------
        context : CalibrationContext
            _description_

        Returns
        -------
        _type_
            _description_
        """
        if custom_range:
            total_neutron_mean = statistics.mean(neutron_counts)
        else:
            all_neutron_values = []
            for day in context.calib_metrics_dict.keys():
                day_neutrons = context.calib_day_df_dict[day][
                    context.neutron_column_name
                ]
                all_neutron_values.extend(day_neutrons.values)

            total_neutron_mean = pd.Series(all_neutron_values).mean()

        n0_range = pd.Series(
            range(int(total_neutron_mean), int(total_neutron_mean * 3.5))
        )
        return n0_range

    def find_optimal_N0(self):
        n0_optimal, rmse = find_n0(
            gravimetric_sm=[
                metrics["field_average_soil_moisture_gravimetric"]
                for metrics in self.context.calib_metrics_dict.values()
            ],
            neutron_count=[
                metrics["average_neutron_count"]
                for metrics in self.context.calib_metrics_dict.values()
            ],
            abs_air_humidity=[
                metrics["absolute_air_humidity"]
                for metrics in self.context.calib_metrics_dict.values()
            ],
            additional_gravimetric_water=self.context.value_avg_lattice_water
            + self.context.value_avg_soil_organic_carbon_water_equiv,
            conversion_theory=self.context.neutron_conversion_method,
            koehli_parameters=self.context.koehli_parameters,
            return_error=True,
        )
        return n0_optimal

    def find_optimal_N0_old(self):
        """
        Finds the optimal N0 number for the site using the weighted
        field average soil mositure.

        Returns
        -------
        average_n0
            The optimal n0 across all the supplied calibration days.
        """
        print("This function is depreciated and will be removed.")
        # Create neutron range
        context = self.context

        if self._using_custom_vals:
            n0_range = self._create_n0_range(
                context=context,
                custom_range=True,
                neutron_counts=self._neutron_counts,
            )
        else:
            n0_range = self._create_n0_range(context=context)
        # print(np.min(n0_range), np.max(n0_range))
        lattice_water = context.value_avg_lattice_water
        water_equiv_soil_organic_carbon = (
            context.value_avg_soil_organic_carbon_water_equiv
        )
        context.calibration_results_by_day = {}

        for day, metrics in context.calib_metrics_dict.items():

            neutron_mean = metrics["average_neutron_count"]
            grav_sm = metrics["field_average_soil_moisture_gravimetric"]

            if context.neutron_conversion_method == "desilets_etal_2010":
                df_calib = self._find_optimal_n0_single_day_desilets_etal_2010(
                    gravimetric_sm_on_day=grav_sm,
                    neutron_mean=neutron_mean,
                    n0_range=n0_range,
                    lattice_water=lattice_water,
                    water_equiv_soil_organic_carbon=water_equiv_soil_organic_carbon,
                )

            elif context.neutron_conversion_method == "koehli_etal_2021":
                # print(grav_sm, neutron_mean, metrics["absolute_air_humidity"], lattice_water, water_equiv_soil_organic_carbon,context.koehli_parameters)
                df_calib = self._find_optimal_n0_single_day_koehli_etal_2021(
                    gravimetric_sm_on_day=grav_sm,
                    neutron_mean=neutron_mean,
                    n0_range=n0_range,
                    abs_air_humidity=metrics["absolute_air_humidity"],
                    lattice_water=lattice_water,
                    water_equiv_soil_organic_carbon=water_equiv_soil_organic_carbon,
                    koehli_parameters=context.koehli_parameters,
                )
            context.calibration_results_by_day[day] = df_calib

        first_day = next(iter(context.calibration_results_by_day))
        total_error_df = pd.DataFrame(
            {"N0": context.calibration_results_by_day[first_day]["N0"]}
        )
        total_error_df["total_error"] = 0
        day_count = 0
        for day_df in context.calibration_results_by_day.values():
            total_error_df["total_error"] += day_df["relative_error"]
            day_count += 1

        new_total_error_col_name = f"total_error_from_{day_count}_calib_days"
        total_error_df.rename(
            {"total_error": new_total_error_col_name},
            axis=1,
            inplace=True,
        )
        min_error_idx = total_error_df[new_total_error_col_name].idxmin()
        n0_optimal = total_error_df.loc[min_error_idx, "N0"]
        return n0_optimal



_LAZY_IMPORTS = {
    "CRNSDataHub": (
        "neptoon.hub",
        "CRNSDataHub",
    ),
}

_ATTR_CACHE = {}


def __getattr__(name: str):
    """Load functions and classes on-demand when first accessed."""

    # Check cache first
    if name in _ATTR_CACHE:
        return _ATTR_CACHE[name]

    if name in _LAZY_IMPORTS:
        module_path, attr_name = _LAZY_IMPORTS[name]

        from importlib import import_module

        module = import_module(module_path)
        attr = getattr(module, attr_name)

        # Cache it for future access
        _ATTR_CACHE[name] = attr
        globals()[name] = attr

        return attr

    raise AttributeError(f"module '{__name__}' has no attribute '{name}'")


def __dir__():
    """Show all available attributes including lazy ones."""
    regular_attrs = [
        name for name in globals().keys() if not name.startswith("_")
    ]
    lazy_attrs = list(_LAZY_IMPORTS.keys())
    return sorted(set(regular_attrs + lazy_attrs))


__all__ = list(_LAZY_IMPORTS.keys())

VERSION = 'v0.13.7'

__version__ = VERSION



# tests/test_pressure_unit_conversion.py
import pandas as pd
import pytest
import datetime
import numpy as np
from neptoon.io.read.data_ingest import (
    InputDataFrameFormattingConfig,
    FormatDataForCRNSDataHub,
    InputColumnMetaData,
    InputColumnDataType,
    PressureUnits,
)
from neptoon.columns import ColumnInfo


@pytest.fixture
def base_config():
    cfg = InputDataFrameFormattingConfig(path_to_config=None)
    return cfg


@pytest.fixture
def base_df():
    df = pd.DataFrame(
        {
            str(ColumnInfo.Name.DATE_TIME): pd.date_range(
                "2022-01-01", periods=5
            ),
        }
    )
    return df


@pytest.fixture
def df_formatter(base_df, base_config):
    df = base_df
    return FormatDataForCRNSDataHub(data_frame=df, config=base_config)


def test_extract_date_time_column(df_formatter):
    series = df_formatter.extract_date_time_column()
    assert isinstance(series, pd.Series)
    assert isinstance(series[0], datetime.datetime)


def test_extract_date_time_column(df_formatter):
    series = df_formatter.extract_date_time_column()
    assert isinstance(series, pd.Series)
    assert isinstance(series[0], datetime.datetime)


def test_extract_date_time_column_list(df_formatter, base_df):
    df_formatter.config.date_time_columns = ["date", "time"]
    df = base_df
    df["date"] = base_df[str(ColumnInfo.Name.DATE_TIME)].dt.date
    df["time"] = base_df[str(ColumnInfo.Name.DATE_TIME)].dt.time
    series = df_formatter.extract_date_time_column()
    assert isinstance(series, pd.Series)
    assert isinstance(series[0], datetime.datetime)


def test_pascals_to_hectopascals(base_config):
    df = pd.DataFrame({"P_raw": [101325, 100000, 98000]})

    base_config.column_data = [
        InputColumnMetaData(
            initial_name="P_raw",
            variable_type=InputColumnDataType.PRESSURE,
            unit=PressureUnits.PASCALS.value,
            priority=1,
        )
    ]
    formatter = FormatDataForCRNSDataHub(
        data_frame=df.copy(), config=base_config
    )
    formatter.standardise_units_of_pressure()
    out = formatter.data_frame

    expected = df["P_raw"] / 100
    pd.testing.assert_series_equal(out["P_raw"], expected, check_names=False)

    assert base_config.column_data[0].unit == PressureUnits.HECTOPASCALS.value


def test_kilopascals_to_hectopascals(base_config):
    df = pd.DataFrame({"P_kpa": [101.325, 100.0, 98.0]})

    base_config.column_data = [
        InputColumnMetaData(
            initial_name="P_kpa",
            variable_type=InputColumnDataType.PRESSURE,
            unit=PressureUnits.KILOPASCALS.value,
            priority=1,
        )
    ]

    formatter = FormatDataForCRNSDataHub(
        data_frame=df.copy(), config=base_config
    )
    formatter.standardise_units_of_pressure()
    out = formatter.data_frame

    expected = df["P_kpa"] * 10
    pd.testing.assert_series_equal(out["P_kpa"], expected, check_names=False)

    assert base_config.column_data[0].unit == PressureUnits.HECTOPASCALS.value


###############
@pytest.fixture
def base_config():
    cfg = InputDataFrameFormattingConfig(path_to_config=None)
    return cfg


@pytest.fixture
def base_df():
    df = pd.DataFrame(
        {
            str(ColumnInfo.Name.DATE_TIME): pd.date_range(
                "2022-01-01", periods=5, freq="1h"
            ),
        }
    )
    return df


@pytest.fixture
def df_formatter(base_df, base_config):
    df = base_df
    return FormatDataForCRNSDataHub(data_frame=df, config=base_config)


def test_conversion_factor_to_cph_one_hour(df_formatter):
    """Test conversion factor calculation for 1-hour timestep."""
    factor = df_formatter.get_conversion_factor_to_cph(3600)
    assert factor == 1.0


def test_conversion_factor_to_cph_thirty_minutes(df_formatter):
    """Test conversion factor calculation for 30-minute timestep."""
    factor = df_formatter.get_conversion_factor_to_cph(1800)
    assert factor == 2.0


def test_conversion_factor_to_cph_one_minute(df_formatter):
    """Test conversion factor calculation for 1-minute timestep."""
    factor = df_formatter.get_conversion_factor_to_cph(60)
    assert factor == 60.0


def test_epi_neutron_counts_per_hour_conversion(base_config):
    """Test conversion when epi neutron data is already in counts_per_hour."""

    epi_col = str(ColumnInfo.Name.EPI_NEUTRON_COUNT_RAW)
    df = pd.DataFrame(
        {
            epi_col: [
                100,
                200,
                150,
                300,
                250,
            ],
            str(ColumnInfo.Name.DATE_TIME): pd.date_range(
                "2022-01-01", periods=5, freq="1h"
            ),
        }
    )

    base_config.column_data = [
        InputColumnMetaData(
            initial_name=epi_col,
            variable_type=InputColumnDataType.EPI_NEUTRON_COUNT,
            unit="counts_per_hour",
            priority=1,
        )
    ]
    base_config.date_time_columns = str(ColumnInfo.Name.DATE_TIME)

    formatter = FormatDataForCRNSDataHub(
        data_frame=df.copy(), config=base_config
    )
    formatter.date_time_as_index()
    formatter.prepare_neutron_count_columns(
        neutron_column_type=InputColumnDataType.EPI_NEUTRON_COUNT
    )

    out = formatter.data_frame

    # Original CPH values should be preserved in the final column
    expected_cph = [100, 200, 150, 300, 250]
    actual_cph = out[str(ColumnInfo.Name.EPI_NEUTRON_COUNT_CPH)].tolist()
    assert actual_cph == expected_cph


def test_epi_neutron_absolute_count_conversion(base_config):
    """Test conversion from absolute_count to counts_per_hour."""
    epi_col = str(ColumnInfo.Name.EPI_NEUTRON_COUNT_RAW)

    df = pd.DataFrame(
        {
            epi_col: [100, 200, 150, 300, 250],
            str(ColumnInfo.Name.DATE_TIME): pd.date_range(
                "2022-01-01", periods=5, freq="1h"
            ),
        }
    )

    base_config.column_data = [
        InputColumnMetaData(
            initial_name="epi_raw",
            variable_type=InputColumnDataType.EPI_NEUTRON_COUNT,
            unit="absolute_count",
            priority=1,
        )
    ]
    base_config.date_time_columns = str(ColumnInfo.Name.DATE_TIME)

    formatter = FormatDataForCRNSDataHub(
        data_frame=df.copy(), config=base_config
    )
    formatter.date_time_as_index()
    formatter.prepare_neutron_count_columns(
        neutron_column_type=InputColumnDataType.EPI_NEUTRON_COUNT
    )

    out = formatter.data_frame

    # For 1-hour timesteps, absolute counts should equal CPH
    # first val will be nan as no diff possible here
    expected_cph = [np.nan, 200, 150, 300, 250]
    out_list = out[str(ColumnInfo.Name.EPI_NEUTRON_COUNT_CPH)].to_list()

    np.testing.assert_array_equal(out_list, expected_cph)


def test_epi_neutron_absolute_count_thirty_minute_timestep(base_config):
    """Test conversion from absolute_count to counts_per_hour with 30-minute timestep."""
    epi_col = str(ColumnInfo.Name.EPI_NEUTRON_COUNT_RAW)
    df = pd.DataFrame(
        {
            epi_col: [50, 100, 75, 150, 125],  # Half-hour absolute counts
            str(ColumnInfo.Name.DATE_TIME): pd.date_range(
                "2022-01-01", periods=5, freq="30min"
            ),
        }
    )

    base_config.column_data = [
        InputColumnMetaData(
            initial_name="epi_raw",
            variable_type=InputColumnDataType.EPI_NEUTRON_COUNT,
            unit="absolute_count",
            priority=1,
        )
    ]
    base_config.date_time_columns = str(ColumnInfo.Name.DATE_TIME)

    formatter = FormatDataForCRNSDataHub(
        data_frame=df.copy(), config=base_config
    )
    formatter.date_time_as_index()
    formatter.prepare_neutron_count_columns(
        neutron_column_type=InputColumnDataType.EPI_NEUTRON_COUNT
    )

    out = formatter.data_frame

    # For 30-minute timesteps, CPH should be double the absolute count
    expected_cph = [np.nan, 200, 150, 300, 250]
    out_list = out[str(ColumnInfo.Name.EPI_NEUTRON_COUNT_CPH)].to_list()

    np.testing.assert_array_equal(out_list, expected_cph)


def test_epi_neutron_counts_per_second_conversion(base_config):
    """Test conversion from counts_per_second to counts_per_hour."""
    epi_col = str(ColumnInfo.Name.EPI_NEUTRON_COUNT_RAW)
    df = pd.DataFrame(
        {
            epi_col: [1.0, 2.0, 0.5, 3.0, 1.5],  # counts per second
            str(ColumnInfo.Name.DATE_TIME): pd.date_range(
                "2022-01-01", periods=5, freq="1h"
            ),
        }
    )

    base_config.column_data = [
        InputColumnMetaData(
            initial_name=epi_col,
            variable_type=InputColumnDataType.EPI_NEUTRON_COUNT,
            unit="counts_per_second",
            priority=1,
        )
    ]
    base_config.date_time_columns = str(ColumnInfo.Name.DATE_TIME)

    formatter = FormatDataForCRNSDataHub(
        data_frame=df.copy(), config=base_config
    )
    formatter.date_time_as_index()
    formatter.prepare_neutron_count_columns(
        neutron_column_type=InputColumnDataType.EPI_NEUTRON_COUNT
    )

    out = formatter.data_frame

    # CPH should be counts_per_second * 3600
    expected_cph = [3600, 7200, 1800, 10800, 5400]
    out_cph = out[str(ColumnInfo.Name.EPI_NEUTRON_COUNT_CPH)].to_list()
    np.testing.assert_array_equal(out_cph, expected_cph)

    # Raw values should be converted to absolute counts (cps * timestep_seconds)
    expected_raw = [np.nan, 7200, 1800, 10800, 5400]
    out_raw = out[str(ColumnInfo.Name.EPI_NEUTRON_COUNT_RAW)].to_list()
    np.testing.assert_array_equal(out_raw, expected_raw)


## REFORMAT FROM HERE


# def test_neutron_uncertainty_calculation(base_config):
#     """Test neutron count uncertainty calculation."""
#     df = pd.DataFrame(
#         {
#             "epi_raw": [
#                 100,
#                 400,
#                 900,
#                 1600,
#             ],  # Perfect squares for easy testing
#             str(ColumnInfo.Name.DATE_TIME): pd.date_range(
#                 "2022-01-01", periods=4, freq="1H"
#             ),
#         }
#     )

#     base_config.column_data = [
#         InputColumnMetaData(
#             initial_name="epi_raw",
#             variable_type=InputColumnDataType.EPI_NEUTRON_COUNT,
#             unit="absolute_count",
#             priority=1,
#         )
#     ]
#     base_config.date_time_columns = str(ColumnInfo.Name.DATE_TIME)

#     formatter = FormatDataForCRNSDataHub(
#         data_frame=df.copy(), config=base_config
#     )
#     formatter.date_time_as_index()
#     formatter.prepare_neutron_count_columns(
#         neutron_column_type=InputColumnDataType.EPI_NEUTRON_COUNT
#     )
#     formatter.calc_neutron_uncertainty()

#     out = formatter.data_frame

#     # Uncertainty should be sqrt of raw counts (for 1-hour timestep, conversion factor = 1)
#     expected_uncertainty = pd.Series(
#         [10, 20, 30, 40],
#         name=str(ColumnInfo.Name.RAW_EPI_NEUTRON_COUNT_UNCERTAINTY),
#     )
#     pd.testing.assert_series_equal(
#         out[str(ColumnInfo.Name.RAW_EPI_NEUTRON_COUNT_UNCERTAINTY)],
#         expected_uncertainty,
#         check_names=False,
#     )


# def test_neutron_uncertainty_with_thirty_minute_timestep(base_config):
#     """Test neutron count uncertainty calculation with 30-minute timestep."""
#     df = pd.DataFrame(
#         {
#             "epi_raw": [100, 400, 900],  # Perfect squares
#             str(ColumnInfo.Name.DATE_TIME): pd.date_range(
#                 "2022-01-01", periods=3, freq="30T"
#             ),
#         }
#     )

#     base_config.column_data = [
#         InputColumnMetaData(
#             initial_name="epi_raw",
#             variable_type=InputColumnDataType.EPI_NEUTRON_COUNT,
#             unit="absolute_count",
#             priority=1,
#         )
#     ]
#     base_config.date_time_columns = str(ColumnInfo.Name.DATE_TIME)

#     formatter = FormatDataForCRNSDataHub(
#         data_frame=df.copy(), config=base_config
#     )
#     formatter.date_time_as_index()
#     formatter.prepare_neutron_count_columns(
#         neutron_column_type=InputColumnDataType.EPI_NEUTRON_COUNT
#     )
#     formatter.calc_neutron_uncertainty()

#     out = formatter.data_frame

#     # Uncertainty should be sqrt(raw_counts) * conversion_factor_to_cph
#     # For 30-minute timestep: conversion_factor = 2.0
#     expected_uncertainty = pd.Series(
#         [20, 40, 60],
#         name=str(ColumnInfo.Name.RAW_EPI_NEUTRON_COUNT_UNCERTAINTY),
#     )
#     pd.testing.assert_series_equal(
#         out[str(ColumnInfo.Name.RAW_EPI_NEUTRON_COUNT_UNCERTAINTY)],
#         expected_uncertainty,
#         check_names=False,
#     )


# def test_multiple_epi_neutron_columns_merge(base_config):
#     """Test merging multiple epithermal neutron columns."""
#     df = pd.DataFrame(
#         {
#             "epi_1": [50, 100, 75, 150, 125],
#             "epi_2": [30, 80, 45, 120, 95],
#             "epi_3": [20, 60, 30, 90, 70],
#             str(ColumnInfo.Name.DATE_TIME): pd.date_range(
#                 "2022-01-01", periods=5, freq="1H"
#             ),
#         }
#     )

#     base_config.column_data = [
#         InputColumnMetaData(
#             initial_name="epi_1",
#             variable_type=InputColumnDataType.EPI_NEUTRON_COUNT,
#             unit="absolute_count",
#             priority=1,
#         ),
#         InputColumnMetaData(
#             initial_name="epi_2",
#             variable_type=InputColumnDataType.EPI_NEUTRON_COUNT,
#             unit="absolute_count",
#             priority=2,
#         ),
#         InputColumnMetaData(
#             initial_name="epi_3",
#             variable_type=InputColumnDataType.EPI_NEUTRON_COUNT,
#             unit="absolute_count",
#             priority=3,
#         ),
#     ]
#     base_config.date_time_columns = str(ColumnInfo.Name.DATE_TIME)

#     formatter = FormatDataForCRNSDataHub(
#         data_frame=df.copy(), config=base_config
#     )
#     formatter.date_time_as_index()
#     formatter.prepare_neutron_count_columns(
#         neutron_column_type=InputColumnDataType.EPI_NEUTRON_COUNT
#     )

#     out = formatter.data_frame

#     # Raw column should be the sum of all epi columns
#     expected_raw = pd.Series(
#         [100, 240, 150, 360, 290],
#         name=str(ColumnInfo.Name.EPI_NEUTRON_COUNT_RAW),
#     )
#     pd.testing.assert_series_equal(
#         out[str(ColumnInfo.Name.EPI_NEUTRON_COUNT_RAW)],
#         expected_raw,
#         check_names=False,
#     )

#     # CPH should equal raw for 1-hour timestep with absolute_count unit
#     pd.testing.assert_series_equal(
#         out[str(ColumnInfo.Name.EPI_NEUTRON_COUNT_CPH)],
#         expected_raw,
#         check_names=False,
#     )


# def test_round_trip_conversion_counts_per_hour_to_absolute_and_back(
#     base_config,
# ):
#     """Test that counts_per_hour conversion is reversible."""
#     original_cph = [100, 200, 300]
#     df = pd.DataFrame(
#         {
#             "epi_raw": original_cph,
#             str(ColumnInfo.Name.DATE_TIME): pd.date_range(
#                 "2022-01-01", periods=3, freq="1H"
#             ),
#         }
#     )

#     base_config.column_data = [
#         InputColumnMetaData(
#             initial_name="epi_raw",
#             variable_type=InputColumnDataType.EPI_NEUTRON_COUNT,
#             unit="counts_per_hour",
#             priority=1,
#         )
#     ]
#     base_config.date_time_columns = str(ColumnInfo.Name.DATE_TIME)

#     formatter = FormatDataForCRNSDataHub(
#         data_frame=df.copy(), config=base_config
#     )
#     formatter.date_time_as_index()
#     formatter.prepare_neutron_count_columns(
#         neutron_column_type=InputColumnDataType.EPI_NEUTRON_COUNT
#     )

#     out = formatter.data_frame

#     # Should recover original CPH values
#     recovered_cph = out[str(ColumnInfo.Name.EPI_NEUTRON_COUNT_CPH)]
#     expected_cph = pd.Series(
#         original_cph, name=str(ColumnInfo.Name.EPI_NEUTRON_COUNT_CPH)
#     )
#     pd.testing.assert_series_equal(
#         recovered_cph, expected_cph, check_names=False
#     )


def test_neutron_uncertainty_calculation(base_config):
    """Test neutron count uncertainty calculation."""
    epi_col = str(ColumnInfo.Name.EPI_NEUTRON_COUNT_RAW)
    df = pd.DataFrame(
        {
            epi_col: [
                100,
                400,
                900,
                1600,
            ],  # Perfect squares for easy testing
            str(ColumnInfo.Name.DATE_TIME): pd.date_range(
                "2022-01-01", periods=4, freq="1h"
            ),
        }
    )

    base_config.column_data = [
        InputColumnMetaData(
            initial_name=epi_col,
            variable_type=InputColumnDataType.EPI_NEUTRON_COUNT,
            unit="absolute_count",
            priority=1,
        )
    ]
    base_config.date_time_columns = str(ColumnInfo.Name.DATE_TIME)

    formatter = FormatDataForCRNSDataHub(
        data_frame=df.copy(), config=base_config
    )
    formatter.date_time_as_index()
    formatter.prepare_neutron_count_columns(
        neutron_column_type=InputColumnDataType.EPI_NEUTRON_COUNT
    )
    formatter.calc_neutron_uncertainty()

    out = formatter.data_frame

    # Uncertainty should be sqrt of raw counts (for 1-hour timestep, conversion factor = 1)
    expected_uncertainty = [np.nan, 20, 30, 40]
    out_uncertainty = out[
        str(ColumnInfo.Name.RAW_EPI_NEUTRON_COUNT_UNCERTAINTY)
    ].to_list()
    np.testing.assert_array_equal(out_uncertainty, expected_uncertainty)


def test_neutron_uncertainty_with_thirty_minute_timestep(base_config):
    """Test neutron count uncertainty calculation with 30-minute timestep."""
    epi_col = str(ColumnInfo.Name.EPI_NEUTRON_COUNT_RAW)
    df = pd.DataFrame(
        {
            epi_col: [100, 400, 900],  # Perfect squares
            str(ColumnInfo.Name.DATE_TIME): pd.date_range(
                "2022-01-01", periods=3, freq="30min"
            ),
        }
    )

    base_config.column_data = [
        InputColumnMetaData(
            initial_name=epi_col,
            variable_type=InputColumnDataType.EPI_NEUTRON_COUNT,
            unit="absolute_count",
            priority=1,
        )
    ]
    base_config.date_time_columns = str(ColumnInfo.Name.DATE_TIME)

    formatter = FormatDataForCRNSDataHub(
        data_frame=df.copy(), config=base_config
    )
    formatter.date_time_as_index()
    formatter.prepare_neutron_count_columns(
        neutron_column_type=InputColumnDataType.EPI_NEUTRON_COUNT
    )
    formatter.calc_neutron_uncertainty()

    out = formatter.data_frame

    # Uncertainty should be sqrt(raw_counts) * conversion_factor_to_cph
    # For 30-minute timestep: conversion_factor = 2.0
    expected_uncertainty = [np.nan, 40, 60]
    out_uncertainty = out[
        str(ColumnInfo.Name.RAW_EPI_NEUTRON_COUNT_UNCERTAINTY)
    ].to_list()
    np.testing.assert_array_equal(out_uncertainty, expected_uncertainty)


def test_multiple_epi_neutron_columns_merge(base_config):
    """Test merging multiple epithermal neutron columns."""
    df = pd.DataFrame(
        {
            "epi_1": [50, 100, 75, 150, 125],
            "epi_2": [30, 80, 45, 120, 95],
            "epi_3": [20, 60, 30, 90, 70],
            str(ColumnInfo.Name.DATE_TIME): pd.date_range(
                "2022-01-01", periods=5, freq="1h"
            ),
        }
    )

    base_config.column_data = [
        InputColumnMetaData(
            initial_name="epi_1",
            variable_type=InputColumnDataType.EPI_NEUTRON_COUNT,
            unit="absolute_count",
            priority=1,
        ),
        InputColumnMetaData(
            initial_name="epi_2",
            variable_type=InputColumnDataType.EPI_NEUTRON_COUNT,
            unit="absolute_count",
            priority=2,
        ),
        InputColumnMetaData(
            initial_name="epi_3",
            variable_type=InputColumnDataType.EPI_NEUTRON_COUNT,
            unit="absolute_count",
            priority=3,
        ),
    ]
    base_config.date_time_columns = str(ColumnInfo.Name.DATE_TIME)

    formatter = FormatDataForCRNSDataHub(
        data_frame=df.copy(), config=base_config
    )
    formatter.date_time_as_index()
    formatter.prepare_neutron_count_columns(
        neutron_column_type=InputColumnDataType.EPI_NEUTRON_COUNT
    )

    out = formatter.data_frame

    # Raw column should be the sum of all epi columns
    expected_raw = [100, 240, 150, 360, 290]
    out_raw = out[str(ColumnInfo.Name.EPI_NEUTRON_COUNT_RAW)].to_list()
    np.testing.assert_array_equal(out_raw, expected_raw)

    # CPH should equal raw for 1-hour timestep with absolute_count unit (first value NaN due to diff)
    expected_cph = [np.nan, 240, 150, 360, 290]
    out_cph = out[str(ColumnInfo.Name.EPI_NEUTRON_COUNT_CPH)].to_list()
    np.testing.assert_array_equal(out_cph, expected_cph)



from pathlib import Path


def is_running_in_docker():
    """
    Checks whether the current kernel is running in a docker container.

    Returns
    -------
    bool
        Whether in docker container
    """
    return Path("/.dockerenv").exists()


def return_file_path_with_suffix(base_path: str):
    """
    This is specifically used internally during docker runs. Input data
    is appended as /workingdir/inputdata. This could be a .csv .zip .tar
    or folder. This will check and return name with correct suffix.

    Parameters
    ----------
    base_path : str
        The path to check

    Returns
    -------
    str
        path with suffix

    """
    base_path = Path(base_path)

    for ext in ["", ".zip", ".tar", ".csv"]:
        test_path = base_path.with_suffix(ext) if ext else base_path
        if test_path.exists():
            actual_input_path = str(test_path)
            break
    else:
        actual_input_path = str(base_path)

    return actual_input_path



import pytest
import platform
from pathlib import Path
from neptoon.utils.general_utils import (
    validate_and_convert_file_path,
    parse_resolution_to_timedelta,
)


def test_file_path_none():
    """Test None Path"""
    tmp = validate_and_convert_file_path(file_path=None)
    assert tmp is None


def test_file_path_str():
    """Test conversion to Path"""
    tmp = validate_and_convert_file_path(file_path="tmp/path")
    assert isinstance(tmp, Path)


def test_file_path_str():
    """Test conversion to Path with base"""
    base = "/base_path/"
    tmp = validate_and_convert_file_path(
        file_path="test/path",
        base=base,
    )
    assert isinstance(tmp, Path)
    assert "base_path" in tmp.parts
    expected = Path(base) / "test/path"
    assert tmp == expected.resolve()


def test_abs_file_path_and_base():
    """Tests AttributeError when a base and an absolute filepath are
    given"""
    base = "/base_path/"
    if platform.system() == "Windows":
        abs_path = "C:/abs/path/"
    else:  # Unix-like (Linux, macOS, etc.)
        abs_path = "/abs/path/"

    with pytest.raises(AttributeError):
        tmp = validate_and_convert_file_path(
            file_path=abs_path,
            base=base,
        )


def test_parse_resolution_good():
    hours = parse_resolution_to_timedelta(resolution_str="1hour")



import pandas as pd
import math
from pathlib import Path
import shutil
import json
import yaml
from typing import List
from magazine import Publish, Magazine
from neptoon.logging import get_logger
from neptoon.config.configuration_input import (
    SensorInfo,
    SensorConfig,
    ProcessConfig,
)
from neptoon.utils import validate_and_convert_file_path
from neptoon.visulisation.figures_handler import FigureHandler
from neptoon.columns import ColumnInfo

core_logger = get_logger()


class SaveAndArchiveOutputs:
    """
    Handles saving the outputs from neptoons in an organised way.

    Future Ideas:
    -------------
    - options to compress outputs (zip_output: bool = True)
    - cloud connection
    - bespoke output formats
    """

    def __init__(
        self,
        folder_name: str,
        processed_data_frame: pd.DataFrame,
        flag_data_frame: pd.DataFrame,
        sensor_info: SensorInfo,
        save_folder_location: str | Path | None = None,
        use_custom_column_names: bool = False,
        custom_column_names_dict: dict | None = None,
        append_timestamp: bool = True,
        figure_handler: FigureHandler | None = None,
        calib_df=None,
        magazine_active: bool = False,
    ):
        """
        Attributes

        Parameters
        ----------
        folder_name : str
            Desired name for the save folder
        processed_data_frame : pd.DataFrame
            The processed time series data
        flag_data_frame : pd.DataFrame
            The flag dataframe
        sensor_info : SensorInfo
            The SensorInfo object.
        save_folder_location : Union[str, Path], optional
            The folder where the data should be saved. If left as None
        use_custom_column_names : bool, optional
             Whether to use custom column names, by default False
        custom_column_names_dict : dict, optional
            A dictionary to convert standard neptoon names into custom a
            custom naming convention, by default None
        append_timestamp: bool, optional, by default True
            Whether to append a timestamp to the folder name when
            saving.
        """
        self.folder_name = folder_name
        self.processed_data_frame = processed_data_frame
        self.flag_data_frame = flag_data_frame
        self.sensor_info = sensor_info
        self.save_folder_location = self._validate_save_folder(
            save_folder_location
        )
        self.use_custom_column_names = use_custom_column_names
        self.custom_column_names_dict = custom_column_names_dict
        self.append_timestamp = append_timestamp
        self.full_folder_location = None
        self.figure_handler = figure_handler
        self.calib_df = calib_df
        self.magazine_active = magazine_active

    def _validate_save_folder(
        self,
        save_location: str | Path | None,
    ):
        """
        Converts string path to pathlib.Path. If given path is not an
        absolute path, saves data to the current working directory.

        Parameters
        ----------
        save_location : Union[str, Path]
            The location where the data should be saved. If a location
            other than the current working directory is desired, provide
            a full path (i.e., not a relative path).

        Returns
        -------
        pathlib.Path
            The pathlib.Path object
        """
        save_path = validate_and_convert_file_path(file_path=save_location)
        if save_path is None:
            save_path = validate_and_convert_file_path(file_path=Path.cwd())
        return save_path

    def create_save_folder(
        self,
    ):
        """
        Creates the folder location where the data will be saved.
        """

        # Make save folder if not already there
        try:
            self.save_folder_location.mkdir()
        except FileExistsError as e:
            message = f"Error: {e} \nFolder already exists."
            core_logger.info(message)

        if self.append_timestamp:
            from datetime import datetime

            # timestamp = datetime.now().strftime("%Y-%m-%d_%H-%M")
            timestamp = datetime.now().strftime("%Y%m%d%H%M%S")
        else:
            timestamp = ""

        new_folder_name = self.folder_name + "_" + timestamp

        self.full_folder_location = self.save_folder_location / new_folder_name

        # Prevent overwriting station data
        try:
            self.full_folder_location.mkdir(parents=True)
        except FileExistsError as e:
            message = f"Error: {e} \nFolder already exists."
            core_logger.error(message)
            print(message + " Please change the folder name and try again.")
            raise FileExistsError

    def close_and_save_data_audit_log(
        self,
        append_hash: bool = False,
    ):
        """
        NOTE: CURRENTLY NOT IMPLEMENTED

        Handles closing the data audit log, producing the YAML output,
        and optionally appending a hash to the save location folder
        name.

        This function performs the following steps:
            1.  Archives and deletes the data audit log using
                DataAuditLog.archive_and_delete_log()

            2. If append_hash is True:
                a. Locates the hash.txt file in the data_audit_log
                   subfolder
                b. Reads the first 6 characters of the hash
                c. Renames the main folder to include this hash

        Parameters:
        -----------
        append_hash : bool, optional (default=False)
            If True, appends the first 6 characters of the hash from
            hash.txt to the folder name.

        """
        from neptoon.data_audit import DataAuditLog

        try:
            DataAuditLog.archive_and_delete_log(
                site_name=self.sensor_info.name,
                custom_log_location=self.full_folder_location,
            )
            if append_hash:
                new_folder_path = self.append_hash_to_folder_name(
                    self.full_folder_location
                )
                # update internal attribute
                self.full_folder_location = new_folder_path
        except AttributeError as e:
            message = f"{e}: DataAuditLog not present - skipping archive step"
            core_logger.info(message)
        except Exception as e:
            message = f"Unexpected error in DataAuditLog archiving: {e}"
            core_logger.error(message)

    def append_hash_to_folder_name(
        self,
        folder_path: Path | None,
    ):
        """
        Appends the first 6 characters of the hash from hash.txt to the
        folder name.

        Parameters:
        -----------
        folder_path : pathlib.Path
            The path to the folder to be renamed.

        Returns:
        --------
        pathlib.Path
            The path to the renamed folder.

        Raises:
        -------
        FileNotFoundError
            If the data audit log folder or hash.txt file is not found.
        PermissionError
            If permissions to access the folder are not available.
        """
        folder_name = folder_path.name
        data_audit_folder = folder_path / "data_audit_log"

        if not data_audit_folder.exists():
            raise FileNotFoundError(
                f"Data audit log folder not found: {data_audit_folder}"
            )

        try:
            unknown_folder_name = next(data_audit_folder.glob("*/"))
            hash_file = unknown_folder_name / "hash.txt"

            if not hash_file.exists():
                raise FileNotFoundError(f"Hash file not found: {hash_file}")

            with hash_file.open("r") as f:
                contents = f.read()

            hash_append = contents[:6]
            new_folder_name = f"{folder_name}_{hash_append}"
            new_folder_path = folder_path.parent / new_folder_name
            folder_path.rename(new_folder_path)

            return new_folder_path

        except StopIteration:
            raise FileNotFoundError(
                f"No subdirectories found in {data_audit_folder}"
            )
        except PermissionError:
            raise PermissionError(
                f"Permission denied when trying to access {folder_path}"
            )
        except Exception as e:
            raise RuntimeError(
                f"An error occurred while appending hash: {str(e)}"
            )

    def mask_bad_data(
        self,
    ):
        """
        Masks out flagged data with nan values
        """
        common_columns = self.flag_data_frame.columns.intersection(
            self.processed_data_frame.columns
        )
        if len(common_columns) < len(self.processed_data_frame.columns):
            core_logger.info(
                "processed_data_frame has additional columns that "
                "will not be masked."
            )
        mask = self.flag_data_frame == "UNFLAGGED"
        masked_df = self.processed_data_frame.copy()
        masked_df[~mask] = math.nan
        return masked_df

    def _save_figures(self):
        """
        Handles saving figures
        """

        figure_metadata = [
            fig_md for fig_md in self.figure_handler.temp_handler.get_figures()
        ]

        figure_folder = self.full_folder_location / "figures"
        figure_folder.mkdir(parents=True, exist_ok=True)
        for figure in figure_metadata:
            try:
                dest = figure_folder / f"{figure.name}.png"
                shutil.copy2(figure.path, dest)
            except FileNotFoundError as err:
                message = f"{figure.name} not found: {err}."
                core_logger.error(message)

    def _update_sensor_info(
        self,
        fields_to_check: List[str] = [
            "beta_coefficient",
            "mean_pressure",
        ],
        beta_col=str(ColumnInfo.Name.BETA_COEFFICIENT),
        mean_press_col=str(ColumnInfo.Name.MEAN_PRESSURE),
    ):
        """
        Updates SensorInfo if values where calulated during processing.

        Parameters
        ----------
        fields_to_check : List[str], optional
            A list of values in SensorInfo to check, by default [
            "beta_coefficient", "l_coefficient", "mean_pressure", ]
        beta_col : str, optional
            Beta Coefficient column name, by default
            str(ColumnInfo.Name.BETA_COEFFICIENT)
        mean_press_col : str, optional
            mean pressure column name, by default
            str(ColumnInfo.Name.MEAN_PRESSURE)
        """
        missing_fields = [
            field
            for field in fields_to_check
            if getattr(self.sensor_info, field) is None
        ]
        if (
            beta_col in self.processed_data_frame.columns
            and "beta_coefficient" in missing_fields
        ):
            beta_coeff = self.processed_data_frame[beta_col].iloc[0]
            self.sensor_info.beta_coefficient = round(beta_coeff, 4)
        if (
            mean_press_col in self.processed_data_frame.columns
            and "mean_pressure" in missing_fields
        ):
            mean_pressure = self.processed_data_frame[mean_press_col].iloc[0]
            self.sensor_info.mean_pressure = round(mean_pressure, 2)

    def _save_pdf(self, location: Path | str):
        """
        Exports the pdf built using magazine to the save folder.

        Parameters
        ----------
        location : Path
            The Path to the folder where the pdf is saved
        """
        if self.magazine_active:
            name = self.sensor_info.name
            save_location = location / f"Report-{name}.pdf"
            with Publish(str(save_location), f"{name} data") as pdf:
                pdf.add_topic("Neutron Correction")
                pdf.add_figure("Neutron Correction")
                pdf.add_topic("NMDB")
                pdf.add_figure("NMDB")
                pdf.add_topic("Soil Moisture")
                pdf.add_figure("Soil Moisture")
                pdf.add_topic("Atmospheric Conditions")
                pdf.add_figure("Atmospheric Conditions")
                pdf.add_topic("Calibration")
                pdf.add_figure("Calibration")
                pdf.add_topic("Data Preparation")
                pdf.add_figure("Data Preparation")
            Magazine.clean()

    def save_data_frames(self, file_name):
        """
        Saves various data frames as .csv files.
        """
        data_folder = self.full_folder_location / "data"
        data_folder.mkdir()
        self.processed_data_frame.to_csv(
            data_folder / f"{file_name}_processed_data.csv"
        )
        self.flag_data_frame.to_csv(data_folder / f"{file_name}_flags.csv")
        if self.calib_df is not None:
            self.calib_df.to_csv(data_folder / f"{file_name}_calibration.csv")

    def save_outputs(
        self,
        use_custom_column_names: bool = False,
    ):
        """
        The main function which chains the options.

        1. Create folder
        2. Mask time series
        3. Save time series
        4. Save flag df
        5. Optional: Save bespoke time series
        6. Optional: Save DAL
        7. Optional: Save Journalist
        8. Optional: rename folder
        9. Optional: compress data
        """
        if use_custom_column_names:
            if self.custom_column_names_dict is None:
                message = (
                    "Cannot use custom column names if no "
                    "column name dictionary supplied."
                )
                core_logger.error(message)
                print(message)
                raise ValueError
        file_name = self.sensor_info.name
        self.create_save_folder()
        self.save_data_frames(file_name=file_name)
        if self.figure_handler:
            self._save_figures()
        self._update_sensor_info()
        if Magazine.active:
            self._save_pdf(location=self.full_folder_location)

    # ---- TODO below this line ----

    def save_custom_column_names(
        self,
    ):
        """
        WIP - save custom variable names using ColumnInfo.
        """
        pass


class ConfigSaver:
    """
    Saves the SensorConfig object as a yaml file.
    """

    def __init__(
        self,
        save_folder_location: Path | str,
        config: SensorConfig,
    ):
        self.save_folder_location = save_folder_location
        self.config = config

    def save(
        self,
    ):
        """
        Convert a Pydantic model to YAML file and save it.
        """
        if isinstance(self.config, SensorConfig):
            save_location = (
                self.save_folder_location / "updated_sensor_config.yaml"
            )
        if isinstance(self.config, ProcessConfig):
            save_location = (
                self.save_folder_location / "updated_process_config.yaml"
            )
        json_str = self.config.model_dump_json()
        data = json.loads(json_str)

        config_str = yaml.safe_dump(
            data,
            default_flow_style=False,
            allow_unicode=True,
            sort_keys=False,
            indent=2,
            default_style=None,
        )

        Path(save_location).write_text(config_str)
